"""
    Implementation of raw hash functions used by multihash multicodecs.

    Hash functions are implemented using the following modules:

    - `hashlib <https://docs.python.org/3/library/hashlib.html>`_, for the ``sha``/``shake`` hash functions and the ``blake2`` hash functions.
    - `pysha3 <https://github.com/tiran/pysha3>`_, for the ``keccak`` hash functions.
    - `blake3 <https://github.com/oconnor663/blake3-py>`_, for the ``blake3`` hash function.
    - `pyskein <https://pythonhosted.org/pyskein/>`_, for the ``skein`` hash functions.
    - `mmh3 <https://github.com/hajimes/mmh3>`_, for the ``murmur3`` hash functions.
    - `pycryptodomex <https://github.com/Legrandin/pycryptodome/>`_, for the ``ripemd-160`` hash function, \
      the ``kangarootwelve`` hash function and the ``sha2-512-224``/``sha2-512-256`` hash functions.

    All modules other than `hashlib <https://docs.python.org/3/library/hashlib.html>`_ are optional dependencies.
    The :func:`get` function attempts to dynamically import any optional dependencies required by desired multihash
    implementation, raising :py:obj:`ImportError` if the dependency is not installed.

    Core functionality is provided by the :func:`exists` and :func:`get` functions,
    which can be used to check whether an implementatino with given name is known, and if so to get the corresponding pair
    of hash function and max digest size:

    >>> multihash.hashfun.exists("sha2-256")
    True
    >>> multihash.hashfun.get("sha2-256")
    (<function _hashlib_sha.<locals>.hashfun at 0x0000013F4A3C6160>, 32)

    The hash functions take a single :obj:`bytes` input (the data) and return a :obj:`bytes` output (the hash digest).
    The max digest sizes (if not :obj:`None`) are used to sense-check hash digests passed to :func:`~multiformats.multihash.wrap`
    and/or obtained from :func:`~multiformats.multihash.unwrap`: telling whether a digest has been generated by a hash function
    is deemed to be computationally unfeasible in general, but hash digests of length greater than the max digest size
    can always be discounted as invalid.
"""

import re
from typing import Dict, Optional, Tuple
from typing_validation import validate

from multiformats import multicodec
from multiformats.varint import BytesLike
from .err import MultihashKeyError, MultihashValueError
from ._hashfuns import Hashfun, validate_hashfun_args, repeat_hashfun

__all__ = ["Hashfun"]

_hashfun: Dict[str, Tuple[Hashfun, Optional[int]]] = {}

MultihashImpl = Tuple[Hashfun, Optional[int]]
"""Type alias for multihash implementations."""

def get(name: str) -> MultihashImpl:
    """
        Given a multihash multicodec name, returns its implementation as a pair of a hash function
        and a max digest size (possibly :obj:`None`).

        >>> multihash.hashfun.get("sha2-256")
        (<function _hashlib_sha.<locals>.hashfun at 0x0000013F4A3C6160>, 32)

        :param name: the name of the multihash
        :type name: :obj:`str`

        :raises KeyError: if no implementation is available for this name

        :rtype: :obj:`MultihashImpl`
    """
    validate(name, str)
    if name not in _hashfun:
        if not _jit_register_hashfun(name):
            raise MultihashKeyError(f"No implementation for multihash multicodec {repr(name)}.")
    return _hashfun[name]


def exists(name: str) -> bool:
    """
        Checks whether the multihash multicodec with given name has an implementation.

        >>> multihash.hashfun.exists("sha2-256")
        True

        :param name: the name of the multihash
        :type name: :obj:`str`

    """
    validate(name, str)
    if name in _hashfun:
        return True
    return _jit_register_hashfun(name, check_only=True)


def register(name: str, hashfun: Hashfun, digest_size: Optional[int], *, overwrite: bool = False) -> None:
    """
        Registers a hash function and hash digest size implementing the multihash multicodec with given name,
        which must already exist.

        Example usage (from the source code of this module):

        .. code-block:: python

            register("sha1", _hashlib_sha(1), 20) # max digest size is 20 bytes, i.e. 160 bits
            register(f"sha2-256", _hashlib_sha(2, 256), 256//8)

        :param name: the name of the multihash
        :type name: :obj:`str`
        :param hashfun: the raw hash function
        :type hashfun: :obj:`Hashfun`
        :param digest_size: the max size for digests, or :obj:`None` if not max size
        :type digest_size: :obj:`int` or :obj:`None`
        :param overwrite: whether an existing implementation with the same name should be overwritten
        :type overwrite: :obj:`bool`, *optional*

        :raises ValueError: if ``overwrite`` is :obj:`False` and an implementation the same name already exists
    """
    validate(name, str)
    # validate(hashfun, Hashfun) # TODO: not yet supported by typing-validation
    validate(digest_size, Optional[int])
    validate(overwrite, bool)
    if digest_size is not None and digest_size <= 0:
        raise MultihashValueError("Digest size must be positive or None.")
    if not overwrite and name in _hashfun:
        raise MultihashValueError(f"An implementation for the multihash multicodec named {repr(name)} already exists.")
    if name not in _hashfun:
        multihash = multicodec.get(name)
        if multihash.tag != "multihash":
            raise MultihashValueError(f"Multicodec '{multihash.name}' exists, but it is not a multihash multicodec.")
    _hashfun[name] = (hashfun, digest_size)


def unregister(name: str) -> None:
    """
        Unregisters a raw encoding by multihash name.

        :param name: the name of the multihash
        :type name: :obj:`str`

        :raises KeyError: if no such raw encoding exists
    """
    validate(name, str)
    if name not in _hashfun:
        raise MultihashKeyError(f"There is no implementation for multihash multicodec with name {repr(name)}.")
    del _hashfun[name]

# identity has function is always registered

def _identity(data: BytesLike, size: Optional[int] = None) -> bytes:
    validate_hashfun_args(data, size, None)
    d = bytes(data)
    if size is None:
        return d
    if len(d) < size:
        raise MultihashValueError("With 'identity' hash, size must be at most data lenght in bytes.")
    return d[:size]

register("identity", _identity, None)

# just-in-time hash implementation registration functions

_sha1_regex = re.compile(r"sha1")
_sha23_regex = re.compile(r"sha(2|3)-(224|256|384|512)")
_shake_regex = re.compile(r"sha(ke)-(128|256)")
_sha2_512_regex = re.compile(r"sha2-512-(224|256)")
_sha2_256_trunc254_padded_regex = re.compile(r"sha2-256-trunc254-padded")

def _jit_register_hashfun_sha(name: str, check_only: bool = False) -> bool:
    # 'sha' hash functions
    m = re.fullmatch(_sha1_regex, name)
    if m is not None:
        from ._hashfuns.sha import _jit_register_sha1 # pylint: disable = import-outside-toplevel
        return _jit_register_sha1(m, None if check_only else register)
    m = re.fullmatch(_sha23_regex, name)
    if m is None:
        m = re.fullmatch(_shake_regex, name)
    if m is not None:
        from ._hashfuns.sha import _jit_register_sha23ke # pylint: disable = import-outside-toplevel
        return _jit_register_sha23ke(m, None if check_only else register)
    m = re.fullmatch(_sha2_512_regex, name)
    if m is not None:
        from ._hashfuns.sha import _jit_register_sha2_512 # pylint: disable = import-outside-toplevel
        return _jit_register_sha2_512(m, None if check_only else register)
    m = re.fullmatch(_sha2_256_trunc254_padded_regex, name)
    if m is not None:
        from ._hashfuns.filecoin import _jit_register_sha_256_trunc254_padded # pylint: disable = import-outside-toplevel
        return _jit_register_sha_256_trunc254_padded(m, None if check_only else register)
    return False

_blake2_regex = re.compile(r"blake2([bs])-([89]|[1-9][0-9]|[1-5][0-9][0-9])")
_blake3_regex = re.compile(r"blake3")

def _jit_register_hashfun_bla(name: str, check_only: bool = False) -> bool:
    # 'blake' hash functions
    m = re.fullmatch(_blake2_regex, name)
    if m is not None:
        from ._hashfuns.blake import _jit_register_blake2 # pylint: disable = import-outside-toplevel
        return _jit_register_blake2(m, None if check_only else register)
    m = re.fullmatch(_blake3_regex, name)
    if m is not None:
        from ._hashfuns.blake import _jit_register_blake3 # pylint: disable = import-outside-toplevel
        return _jit_register_blake3(m, None if check_only else register)
    return False

_keccak_regex = re.compile(r"keccak-(224|256|384|512)")

def _jit_register_hashfun_kec(name: str, check_only: bool = False) -> bool:
    # 'keccak' hash function
    m = re.fullmatch(_keccak_regex, name)
    if m is not None:
        from ._hashfuns.keccak import _jit_register_keccak # pylint: disable = import-outside-toplevel
        return _jit_register_keccak(m, None if check_only else register)
    return False

_skein_regex = re.compile(r"skein(256|512|1024)-([89]|[1-9][0-9]|[1-9][0-9][0-9]|10[0-2][0-9])")

def _jit_register_hashfun_ske(name: str, check_only: bool = False) -> bool:
    # 'skein' hash function
    m = re.fullmatch(_skein_regex, name)
    if m is not None:
        from ._hashfuns.skein import _jit_register_skein # pylint: disable = import-outside-toplevel
        return _jit_register_skein(m, None if check_only else register)
    return False

_murmur3_regex = re.compile(r"murmur3-(32)|murmur3-(x64)-(64|128)")

def _jit_register_hashfun_mur(name: str, check_only: bool = False) -> bool:
    # 'murmur3' hash function
    m = re.fullmatch(_murmur3_regex, name)
    if m is not None:
        from ._hashfuns.murmur3 import _jit_register_murmur3 # pylint: disable = import-outside-toplevel
        return _jit_register_murmur3(m, None if check_only else register)
    return False

_md5_regex = re.compile("md5")

def _jit_register_hashfun_md5(name: str, check_only: bool = False) -> bool:
    # 'md5' hash function
    m = re.fullmatch(_md5_regex, name)
    if m is not None:
        from ._hashfuns.md import _jit_register_md5 # pylint: disable = import-outside-toplevel
        return _jit_register_md5(m, None if check_only else register)
    return False

_ripemd_regex = re.compile(r"ripemd-(160)")

def _jit_register_hashfun_rip(name: str, check_only: bool = False) -> bool:
    # 'ripemd' hash functions
    m = re.fullmatch(_ripemd_regex, name)
    if m is not None:
        from ._hashfuns.md import _jit_register_ripemd # pylint: disable = import-outside-toplevel
        return _jit_register_ripemd(m, None if check_only else register)
    return False

_kangarootwelve_regex = re.compile("kangarootwelve")

def _jit_register_hashfun_kan(name: str, check_only: bool = False) -> bool:
    # 'kangarootwelve' hash function
    m = re.fullmatch(_kangarootwelve_regex, name)
    if m is not None:
        from ._hashfuns.kangarootwelve import _jit_register_kangarootwelve # pylint: disable = import-outside-toplevel
        return _jit_register_kangarootwelve(m, None if check_only else register)
    return False

_dbl_sha2_regex = re.compile(r"dbl-sha2-(256)")

def _jit_register_hashfun_dbl(name: str, check_only: bool = False) -> bool:
    # 'dbl-sha2-256' hash function
    m = re.fullmatch(_dbl_sha2_regex, name)
    if m is not None:
        sha2_256, _ = get("sha2-256")
        assert sha2_256 is not None
        dbl_sha2_256 = repeat_hashfun(sha2_256, repeat=2, truncate="end")
        register("dbl-sha2-256", dbl_sha2_256, 32)
        return True
    return False

# directory of just-in-time hash implementation registration functions

_jit_register_hashfun_dir = {
    "sha": _jit_register_hashfun_sha,
    "bla": _jit_register_hashfun_bla,
    "kec": _jit_register_hashfun_kec,
    "ske": _jit_register_hashfun_ske,
    "mur": _jit_register_hashfun_mur,
    "md5": _jit_register_hashfun_md5,
    "rip": _jit_register_hashfun_rip,
    "kan": _jit_register_hashfun_kan,
    "dbl": _jit_register_hashfun_dbl,
}

def _jit_register_hashfun(name: str, check_only: bool = False) -> bool:
    # pylint: disable = too-many-return-statements
    if len(name) < 3:
        return False
    jit_reg_fun = _jit_register_hashfun_dir.get(name[:3], None)
    if jit_reg_fun is None:
        return False
    return jit_reg_fun(name, check_only)



"""
    Errors for the :mod:`~multiformats.multibase` module.
"""

import builtins

class MultibaseKeyError(builtins.KeyError): # pylint: disable = redefined-builtin
    """ Class for :mod:`~multiformats.multibase` key errors. """

class MultibaseValueError(builtins.ValueError): # pylint: disable = redefined-builtin
    """ Class :mod:`~multiformats.multibase` value errors. """



"""
    Implementation of the `multihash spec <https://github.com/multiformats/multihash>`_.

    Suggested usage:

    >>> from multiformats import multihash
"""

from io import BytesIO, BufferedIOBase
from typing import AbstractSet, Any, cast, ClassVar, Dict, Iterator, Mapping, Optional, overload, Union, Sequence, Tuple, Type, TypeVar
from weakref import WeakValueDictionary
import sys
from typing_extensions import Literal
from typing_validation import validate

from multiformats import multicodec, varint
from multiformats.multicodec import Multicodec, _hexcode
from multiformats.varint import BytesLike

from . import raw
from .raw import Hashfun, MultihashImpl
from .err import MultihashKeyError, MultihashValueError

class Multihash:
    """
        Container class for a multibase encoding.

        Example usage:

        >>> sha2_256 = multihash.get("sha2-256")
        >>> sha2_256
        Multihash(codec='sha2-256')

        :param codec: the multicodec defining this multihash
        :type codec: :obj:`str`, :obj:`int` or :class:`~multiformats.multicodec.Multicodec`
    """

    # WeakValueDictionary[str, Multihash]
    _cache: ClassVar[WeakValueDictionary] = WeakValueDictionary() # type: ignore

    _codec: Multicodec
    _implementation: Optional[MultihashImpl]

    __slots__ = ("__weakref__", "_codec", "_implementation")

    def __new__(cls, codec: Union[str, int, Multicodec]) -> "Multihash":
        # check that the codec exists:
        if isinstance(codec, str):
            codec = multicodec.get(codec)
        elif isinstance(codec, int):
            codec = multicodec.get(code=codec)
        else:
            validate(codec, Multicodec)
            existing_codec = multicodec.get(codec.name)
            if existing_codec != codec:
                raise MultihashValueError(f"Multicodec named {repr(codec.name)} exists, but is not the one given.")
            codec = existing_codec
        # check that the codec is a multihash multicodec:
        if codec.tag != "multihash":
            raise MultihashValueError(f"Multicodec named {repr(codec.name)} exists, but is not a multihash.")
        if not raw.exists(codec.name):
            raise MultihashKeyError(f"No implementation for multihash multicodec {repr(codec.name)}.")
        _cache = Multihash._cache
        if codec.name in _cache:
            # if a multihash instance with this name is already registered
            instance: Multihash = _cache[codec.name]
            if instance.codec == codec:
                # same codec, check same implementation:
                if instance._implementation is None:
                    # implementation not loaded yet, can use the existing instance
                    return instance
                if codec.name in raw._hashfun and instance._implementation == raw._hashfun[codec.name]:
                    # nothing changed, can use the existing instance
                    return instance
            # otherwise remove the existing instance
            del _cache[codec.name]
        # create a fresh instance, register it and return it
        instance = super().__new__(cls)
        instance._codec = codec
        instance._implementation = None
        _cache[codec.name] = instance
        return instance

    @property
    def name(self) -> str:
        """
            Multihash multicodec name.

            Example usage:

            >>> sha2_256.name
            'sha2-256'

        """
        return self.codec.name

    @property
    def code(self) -> int:
        """
            Multihash multicodec code.

            Example usage:

            >>> sha2_256.code
            18
            # 18 = 0x12

        """
        return self.codec.code

    @property
    def codec(self) -> Multicodec:
        """
            The multicodec for this multihash.

            Example usage:

            >>> sha2_256.codec
            Multicodec(name='sha2-256', tag='multihash', code='0x12',
                       status='permanent', description='')

        """
        return self._codec

    @property
    def max_digest_size(self) -> Optional[int]:
        """
            The maximum size (in bytes) for raw digests of this multihash, or :obj:`None` if there is no maximum size.
            Used to sense-check the wrapped/unwrapped raw digests.

            Example usage:

            >>> sha2_256.max_digest_size
            32
            # 32 bytes = 256 bits

        """
        _, max_digest_size = self.implementation
        return max_digest_size

    @property
    def implementation(self) -> MultihashImpl:
        """
            Returns the implementation of a multihash multicodec, as a pair:

            .. code-block:: python

                hash_function, max_digest_size = multihash.implementation("sha2-256")

            Above, ``codec`` is the :class:`~multiformats.multicodec.Multicodec` object carrying information about the
            multihash multicodec, ``hash_function`` is the function `bytes->bytes` computing the raw hashes,
            and ``max_digest_size`` is the max size of the digests produced by ``hash_function`` (or :obj:`None` if
            there is no max size, such as in the case of the ``'identity'`` multihash multicodec).

            Example usage:

            >>> sha2_256.implementation
            (<function _hashlib_sha.<locals>.hashfun at 0x0000029396E22280>, 32)

            :rtype: :obj:`~multiformats.multihash.raw.MultihashImpl`

        """
        implementation = self._implementation
        if implementation is None:
            implementation = raw.get(self.name)
            self._implementation = implementation
        return implementation

    def wrap(self, raw_digest: BytesLike) -> bytes:
        """
            Wraps a raw digest into a multihash digest:

            .. code-block::

                <raw digest> --> <code><size><raw digest>

            Example usage:

            >>> sha2_256 = multihash.get("sha2-256")
            >>> raw_digest = bytes.fromhex(
            ... "c0535e4be2b79ffd93291305436bf889314e4a3f")
            >>> sha2_256.wrap(raw_digest).hex()
            "1214c0535e4be2b79ffd93291305436bf889314e4a3f"

            :param raw_digest: the raw digest
            :type raw_digest: :obj:`~multiformats.varint.BytesLike`

            See :func:`wrap` for more information.
        """
        validate(raw_digest, BytesLike)
        _, max_digest_size = self.implementation
        size = len(raw_digest)
        if max_digest_size is not None and size > max_digest_size:
            raise MultihashValueError(f"Digest size {max_digest_size} is listed for {self.name}, "
                             f"but a digest of larger size {size} was given to be wrapped.")
        return self.codec.wrap(varint.encode(size)+raw_digest)

    def digest(self, data: BytesLike, *, size: Optional[int] = None) -> bytes:
        """
            Computes the raw digest of the given data and wraps it into a multihash digest.

            Example usage:

            >>> sha2_256 = multihash.get("sha2-256")
            >>> data = b"Hello world!"
            >>> data.hex()
            "48656c6c6f20776f726c6421"
            >>> sha2_256.digest(data).hex() # full 32-bytes hash
            '1220c0535e4be2b79ffd93291305436bf889314e4a3faec05ecffcbb7df31ad9e51a'
            >>> sha2_256.digest(data, size=20).hex() # truncated hash
            '1214c0535e4be2b79ffd93291305436bf889314e4a3f'

            :param data: the raw digest
            :type data: :obj:`~multiformats.varint.BytesLike`
            :param size: size for the raw digest, in bytes. If not :obj:`None`, raw digest is truncated to fit the given size.
            :type size: :obj:`int` or :obj:`None`, *optional*

            :raises ValueError: if size parameter is not :obj:`None` and negative.
            :raises ValueError: if size parameter is not :obj:`None`, max digest size is not :obj:`None` and given size exceeds max digest size.
            :raises ValueError: if size parameter is :obj:`None` but a size is required for the hash function (e.g. for the KangarooTwelve XOF).

            See :func:`digest` for more information.
        """
        hf, _ = self.implementation
        raw_digest = hf(data, size)
        # if size is not None:
        #     raw_digest = raw_digest[:size] # truncate digest
        if size is None:
            size = len(raw_digest)
        else:
            assert size == len(raw_digest), f"Expected {size}B digest, found {len(raw_digest)}B digest."
        return self.codec.wrap(varint.encode(size)+raw_digest)

    def unwrap(self, digest: Union[BytesLike, BufferedIOBase]) -> bytes:
        """
            Unwraps a multihash digest into a hash digest:

            .. code-block::

                <code><size><raw digest> --> <raw digest>

            If `digest` is one or :obj:`bytes`, :obj:`bytearray` or :obj:`memoryview`, the method also checks
            that the actual hash digest size matches the size listed by the multihash digest.

            Example usage:

            >>> sha2_256 = multihash.get("sha2-256")
            >>> digest = bytes.fromhex(
            ... "1214c0535e4be2b79ffd93291305436bf889314e4a3f")
            >>> sha2_256.unwrap(digest).hex()
            'c0535e4be2b79ffd93291305436bf889314e4a3f'

            :param digest: the multihash digest to be unwrapped
            :type digest: :obj:`~multiformats.varint.BytesLike` or :obj:`~io.BufferedIOBase`

            See :func:`unwrap` for more information.

        """
        code, raw_digest = unwrap_raw(digest)
        if code != self.code:
            raise MultihashValueError(f"Decoded code {code} differs from multihash code {self.code}.")
        _validate_raw_digest_size(self.name, raw_digest, self.max_digest_size)
        return raw_digest

    def __str__(self) -> str:
        return f"multihash.get({repr(self.name)})"

    def __repr__(self) -> str:
        return f"Multihash(codec={repr(self.name)})"

    @property
    def _as_tuple(self) -> Tuple[Type["Multihash"], Multicodec]:
        return (Multihash, self.codec)

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, Multihash):
            return NotImplemented
        return self._as_tuple == other._as_tuple


def get(name: Optional[str] = None, *, code: Optional[int] = None) -> Multihash:
    """
        Gets the multihash multicodec with given name or code.

        Example usage:

        >>> multihash.get("sha1")
        Multihash(codec='sha1')
        >>> multihash.get(code=0x11)
        Multihash(codec='sha1')

        :param name: the multihash name
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the multihash code
        :type code: :obj:`int` or :obj:`None`, *optional*

        :raises KeyError: if the multihash does not exist or is not implemented
        :raises ValueError: unless exactly one of ``name`` and ``code`` is specified

    """
    if name is not None and code is not None:
        raise MultihashValueError("Must specify at most one between 'name' and 'code'.")
    if name is not None:
        return Multihash(codec=name)
    if code is not None:
        return Multihash(codec=code)
    raise MultihashValueError("Must specify at least one between 'name' and 'code'.")


def exists(name: Optional[str] = None, *, code: Optional[int] = None) -> bool:
    """
        Checks whether a multihash multicodec with the given name or code exists.
        This function returns `False` if a multicodec by given name or code exists, but is not tagged ``'multihash'``.

        Example usage:

        >>> multihash.exists("sha1")
        True
        >>> multihash.exists(code=0x11)
        True
        >>> from multiformats import multicodec
        >>> multicodec.get("cidv1")
        Multicodec(name='cidv1', tag='cid', code='0x01',
                   status='permanent', description='CIDv1')
        >>> multihash.exists("cidv1")
        False

        :param name: the multihash name
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the multihash code
        :type code: :obj:`int` or :obj:`None`, *optional*

        :raises ValueError: unless exactly one of ``name`` and ``code`` is specified

    """
    if not multicodec.exists(name, code=code):
        return False
    codec = multicodec.get(name, code=code)
    return codec.tag == "multihash"


def is_implemented(name: Optional[str] = None, *, code: Optional[int] = None) -> bool:
    """
        Checks whether a multihash with the given name or code exists and is implemented.

        Example usage:

        >>> multihash.is_implemented("sha1")
        True
        >>> multihash.is_implemented(code=0x11)
        True

        :param name: the multihash name
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the multihash code
        :type code: :obj:`int` or :obj:`None`, *optional*

        :raises ValueError: unless exactly one of ``name`` and ``code`` is specified
    """
    if not exists(name, code=code):
        return False
    codec = multicodec.get(name, code=code)
    return raw.exists(codec.name)


def from_digest(digest: BytesLike) -> Multihash:
    """
        Returns the multihash multicodec for the given digest,
        according to the code specified by its prefix.

        Example usage:

        >>> digest = bytes.fromhex("140a9a7a8207a57d03e9c524")
        >>> multihash.from_digest(digest)
        Multihash(codec='sha3-512')

        :param digest: the multihash digest
        :type digest: :obj:`~multiformats.varint.BytesLike`

        :raises KeyError: if no multihash exists with that code
    """
    code, _, _ = multicodec.unwrap_raw(digest)
    return get(code=code)


def wrap(raw_digest: BytesLike, hashfun: Union[str, int, Multihash]) -> bytes:
    """
        Wraps a raw digest into a multihash digest using the given multihash:

        .. code-block::

            <raw digest> --> <code><size><raw digest>

        If the multihash is passed by name or code, the :func:`get` function is used to retrieve it.

        Example usage:

        >>> multihash.get("sha2-256").codec
        Multicodec(name='sha2-256', tag='multihash', code='0x12',
                   status='permanent', description='')
        >>> raw_digest = bytes.fromhex("c0535e4be2b79ffd93291305436bf889314e4a3f")
        >>> len(raw_digest)
        20
        >>> multihash.wrap(raw_digest, "sha2-256").hex()
        "1214c0535e4be2b79ffd93291305436bf889314e4a3f"
        #^^   code 0x12 for multihash multicodec "sha2-256"
        #  ^^ truncated hash length 0x14 = 20 bytes

        Note that all digests above are :obj:`bytes` objects, represented here as hex strings for clarity:

        >>> hash_digest
        b'\\xc0S^K\\xe2\\xb7\\x9f\\xfd\\x93)\\x13\\x05Ck\\xf8\\x891NJ?'
        >>> multihash.wrap(raw_digest, "sha2-256")
        b'\\x12\\x14\\xc0S^K\\xe2\\xb7\\x9f\\xfd\\x93)\\x13\\x05Ck\\xf8\\x891NJ?'
        # ^^^^     0x12 -> multihash multicodec "sha2-256"
        #     ^^^^ 0x14 -> truncated hash length of 20 bytes

        :param raw_digest: the raw hash digest
        :type raw_digest: :obj:`~multiformats.varint.BytesLike`
        :param hashfun: the multihash function name, code or object
        :type hashfun: :obj:`str`, :obj:`int` or :class:`Multihash`

    """
    if not isinstance(hashfun, Multihash):
        hashfun = Multihash(codec=hashfun)
    return hashfun.wrap(raw_digest)


def digest(data: BytesLike, hashfun: Union[str, int, Multihash], *, size: Optional[int] = None) -> bytes:
    """
        Computes the raw digest of the given data and wraps it into a multihash digest.

        Example usage:

        >>> data = b"Hello world!"
        >>> data.hex()
        "48656c6c6f20776f726c6421"
        >>> multihash.digest(data, "sha2-256").hex() # full 32-bytes hash
        '1220c0535e4be2b79ffd93291305436bf889314e4a3faec05ecffcbb7df31ad9e51a'
        >>> multihash.digest(data, "sha2-256", size=20).hex()
        #         optional truncated hash size ^^^^^^^
        '1214c0535e4be2b79ffd93291305436bf889314e4a3f'
        #^^   code 0x12 for multihash multicodec "sha2-256"
        #  ^^ truncated hash length 0x14 = 20 bytes

        Note both multihash code and digest length are wrapped as varints
        (see the `multiformats.varint` module) and can span multiple bytes:

        >>> multihash.get("skein1024-1024")
        Multicodec(name='skein1024-1024', tag='multihash', code='0xb3e0',
                   status='draft', description='')
        >>> multihash.digest(data, "skein1024-1024").hex()
        'e0e702800192e08f5143 ... 3+2+128 = 133 bytes in total
        #^^^^^^     3-bytes varint for hash function code 0xb3e0
        #      ^^^^ 2-bytes varint for hash digest length 128
        >>> from multiformats import varint
        >>> hex(varint.decode(bytes.fromhex("e0e702")))
        '0xb3e0'
        >>> varint.decode(bytes.fromhex("8001"))
        128

        :param data: the data to be digested
        :type data: :obj:`~multiformats.varint.BytesLike`
        :param hashfun: the multihash function name, code or object
        :type hashfun: :obj:`str`, :obj:`int` or :class:`Multihash`
        :param size: size for the raw digest, in bytes. If not :obj:`None`, raw digest is truncated to fit the given size.
        :type size: :obj:`int` or :obj:`None`, *optional*

        :raises ValueError: if size parameter is not :obj:`None` and negative.
        :raises ValueError: if size parameter is not :obj:`None`, max digest size is not :obj:`None` and given size exceeds max digest size.
        :raises ValueError: if size parameter is :obj:`None` but a size is required for the hash function (e.g. for the KangarooTwelve XOF).

    """
    if not isinstance(hashfun, Multihash):
        hashfun = Multihash(codec=hashfun)
    return hashfun.digest(data, size=size)


def _validate_raw_digest_size(name: str, raw_digest: bytes, max_digest_size: Optional[int]) -> None:
    if max_digest_size is not None and len(raw_digest) > max_digest_size:
        raise MultihashValueError(f"Multihash {name} has max digest size {max_digest_size}, "
                         f"but a digest of larger size {len(raw_digest)} was unwrapped instead.")


def unwrap(digest: Union[BytesLike, BufferedIOBase],
           hashfun: Union[None, str, int, Multihash]=None) -> bytes:
    """
        Unwraps a multihash digest into a raw digest:

        .. code-block::

            <code><size><raw digest> --> <raw digest>

        If ``digest`` is one of :obj:`bytes`, :obj:`bytearray` or :obj:`memoryview`, the method also checks
        that the actual raw digest size matches the size listed in the multihash digest.
        If ``digest`` is a stream (an instance of :obj:`~io.BufferedIOBase`, specifically), then the
        number of bytes consumed to produce the raw digest matches the size lised in the multihash digest,
        and no further bytes are consumed from the stream.

        If the optional ``hashfun`` is not ``None``, the function additionally enforces that the code from the
        multihash digest matches the code of the multihash (calls :func:`Multihash.unwrap`
        under the hood to do so). Regardless, the function checks that the multihash with code specified by the
        multihash digest exists and is implemented.

        Example usage:

        >>> digest = bytes.fromhex(
        ... "1214c0535e4be2b79ffd93291305436bf889314e4a3f")
        >>> multihash.unwrap(digest, "sha2-256").hex()
        'c0535e4be2b79ffd93291305436bf889314e4a3f'

        :param digest: the multihash digest
        :type digest: :obj:`~multiformats.varint.BytesLike` or :obj:`~io.BufferedIOBase`
        :param hashfun: the multihash function name, code or object
        :type hashfun: :obj:`str`, :obj:`int` or :class:`Multihash`, *optional*

        :raises ValueError: if the unwrapped raw digest exceeds the stated maximum digest size for the multihash function

    """
    if hashfun is not None:
        if not isinstance(hashfun, Multihash):
            hashfun = Multihash(codec=hashfun)
        return hashfun.unwrap(digest)
    code, raw_digest = unwrap_raw(digest)
    hashfun = Multihash(codec=code)
    _validate_raw_digest_size(hashfun.name, raw_digest, hashfun.max_digest_size)
    return raw_digest


_BufferedIOT = TypeVar("_BufferedIOT", bound=BufferedIOBase)

@overload
def unwrap_raw(digest: BufferedIOBase) -> Tuple[int, bytes]:
    ...

@overload
def unwrap_raw(digest: BytesLike) -> Tuple[int, memoryview]:
    ...

def unwrap_raw(digest: Union[BytesLike, BufferedIOBase]) -> Tuple[int, Union[bytes, memoryview]]:
    """
        Unwraps a multihash digest into a code and raw digest pair:

        .. code-block::

            <code><size><raw digest> --> (<code>, <raw digest>)

        The function checks that the multihash codec with code specified by the multihash digest exists,
        but does not check whether it is implemented or not.

        Example usage:

        >>> digest = bytes.fromhex(
        ... "1214c0535e4be2b79ffd93291305436bf889314e4a3f")
        >>> code, digest = multihash.unwrap_raw(digest, "sha2-256")
        >>> code
        18 # the code 0x12 of 'sha2-256'
        >>> digest.hex()
        'c0535e4be2b79ffd93291305436bf889314e4a3f'

        :param digest: the multihash digest
        :type digest: :obj:`~multiformats.varint.BytesLike` or :obj:`~io.BufferedIOBase`

    """
    # switch between memoryview mode and stream mode
    if isinstance(digest, BufferedIOBase):
        stream_mode = True
        validate(digest, BufferedIOBase)
        stream: Union[memoryview, BufferedIOBase] = digest
    else:
        stream_mode = False
        stream = memoryview(digest)
    # extract multihash code
    code, n, stream = multicodec.unwrap_raw(digest)
    if not exists(code=code):
        n_bytes_read = f" ({n} bytes read)" if stream_mode else ""
        raise MultihashKeyError(f"Multicodec {_hexcode(code)} is not a multihash{n_bytes_read}.")
    # extract hash digest size
    digest_size, _, stream = varint.decode_raw(stream)
    # extract hash digest
    if stream_mode:
        # use only the number of bytes specified by the multihash
        hash_digest = cast(BufferedIOBase, stream).read(digest_size)
    else:
        # use all remaining bytes
        hash_digest = cast(memoryview, stream)
    # check that the hash digest size is valid
    if digest_size != len(hash_digest):
        raise MultihashValueError(f"Multihash digest lists size {digest_size}, but the hash digest has size {len(hash_digest)} instead.")
    return code, hash_digest



"""
    A script to generate .rst files for API documentation.
"""

import glob
import importlib
import inspect
import json
import os
import pkgutil
from typing import Dict, List, Optional, Tuple
import sys

from typing_validation import validate

def _list_package_contents(pkg_name: str) -> List[str]:
    modules = [pkg_name]
    for submod in pkgutil.iter_modules([pkg_name.replace(".", "/")]):
        submod_fullname = pkg_name+"."+submod.name
        if submod.ispkg:
            for subsubmod_name in _list_package_contents(submod_fullname):
                modules.append(subsubmod_name)
        else:
            modules.append(submod_fullname)
    return modules

def make_apidocs() -> None:
    """
        A script to generate .rst files for API documentation.
    """
    err_msg = """Expected a 'make-api.json' file, with the following structure:
{
    "pkg_name": str,
    "apidocs_folder": str,
    "pkg_path": str,
    "toc_filename": Optional[str],
    "include_members": Dict[str, List[str]],
    "exclude_members": Dict[str, List[str]],
    "exclude_modules": List[str]
}

Set "toc_filename" to null to avoid generating a table of contents file.

"""
    try:
        with open("make-api.json", "r") as f:
            config = json.load(f)
            pkg_name = config.get("pkg_name", None)
            validate(pkg_name, str)
            pkg_path = config.get("pkg_path", None)
            validate(pkg_path, str)
            apidocs_folder = config.get("apidocs_folder", None)
            validate(apidocs_folder, str)
            toc_filename = config.get("toc_filename", None)
            validate(toc_filename, Optional[str])
            include_members = config.get("include_members", None)
            validate(include_members, Dict[str, List[str]])
            exclude_members = config.get("exclude_members", None)
            validate(exclude_members, Dict[str, List[str]])
            include_modules = config.get("include_modules", None)
            validate(include_modules, List[str])
            exclude_modules = config.get("exclude_modules", None)
            validate(exclude_modules, List[str])
    except FileNotFoundError:
        print(err_msg)
        sys.exit(1)
    except TypeError:
        print(err_msg)
        sys.exit(1)

    cwd = os.getcwd()
    os.chdir(pkg_path)
    sys.path = [os.getcwd()]+sys.path
    modules = _list_package_contents(pkg_name)
    modules_dict = {
        mod_name: importlib.import_module(mod_name)
        for mod_name in modules
    }
    for mod_name in include_modules:
        if mod_name not in modules_dict:
            modules_dict[mod_name] = importlib.import_module(mod_name)
    os.chdir(cwd)

    print(f"Removing all docfiles from {apidocs_folder}/")
    for apidoc_file in glob.glob(f"{apidocs_folder}/*.rst"):
        print(f"    {apidoc_file}")
        os.remove(apidoc_file)
    print()

    mod_name_to_del: List[str] = []

    for mod_name, mod in modules_dict.items():
        if any(mod_name.startswith(name) for name in exclude_modules):
        # if mod_name in exclude_modules:
            mod_name_to_del.append(mod_name)
            continue
        filename = f"{apidocs_folder}/{mod_name}.rst"
        print(f"Writing API docfile {filename}")
        lines: List[str] = [
            mod_name,
            "="*len(mod_name),
            "",
            f".. automodule:: {mod_name}",
            ""
        ]
        mod__all__ = getattr(mod, "__all__", [])
        reexported_members: List[Tuple[str, str]] = []
        for member_name in sorted(name for name in dir(mod) if not name.startswith("_")):
            if mod_name in exclude_members and member_name in exclude_members[mod_name]:
                continue
            member = getattr(mod, member_name)
            member_module = inspect.getmodule(member)
            member_module_name = member_module.__name__ if member_module is not None else None
            imported_member = member_module is not None and member_module != mod
            if mod_name in include_members and member_name in include_members[mod_name]:
                imported_member = False
            if imported_member:
                if inspect.ismodule(member):
                    member_fullname = member_module_name or ""
                else:
                    member_fullname = f"{member_module_name}.{member_name}"
            else:
                member_fullname = f"{mod_name}.{member_name}"
            member_kind = "data"
            if inspect.isclass(member):
                member_kind = "class"
            elif inspect.isfunction(member):
                member_kind = "function"
            elif inspect.ismodule(member):
                member_kind = "module"
            if not imported_member:
                member_lines: List[str] = []
                member_lines = [
                    member_name,
                    "-"*len(member_name),
                    "",
                    f".. auto{member_kind}:: {member_fullname}",
                ]
                if member_kind == "class":
                    member_lines.append("    :members:")
                member_lines.append("")
                print(f"    {member_kind} {member_name}")
                lines.extend(member_lines)
            elif member_name in mod__all__:
                reexported_members.append((member_fullname, member_kind))
        if reexported_members:
            reexported_members_header = f"{mod_name}.__all__"
            print(f"    {reexported_members_header}:")
            lines.extend([
                reexported_members_header,
                "-"*len(reexported_members_header),
                "",
                "The following members were explicitly reexported using ``__all__``:",
                "",
            ])
            refkinds = {
                "data": "obj",
                "function": "func",
                "class": "class",
                "module": "mod"
            }
            for member_fullname, member_kind in reexported_members:
                refkind = f":py:{refkinds[member_kind]}:"
                lines.append(f"    - {refkind}`{member_fullname}`")
                print(f"        {member_kind} {member_fullname}")
            lines.append("")
        with open(filename, "w") as f:
            f.write("\n".join(lines))
        print("")


    for mod_name in mod_name_to_del:
        del modules_dict[mod_name]

    toctable_lines = [
        ".. toctree::",
        "    :maxdepth: 2",
        "    :caption: API Documentation",
        ""
    ]
    print(f"Writing TOC for API docfiles at {toc_filename}")
    for mod_name in modules_dict:
        line = f"    {apidocs_folder}/{mod_name}"
        toctable_lines.append(line)
        print(line)
    toctable_lines.append("")
    print()

    with open(toc_filename, "w") as f:
        f.write("\n".join(toctable_lines))

if __name__ == "__main__":
    make_apidocs()



"""
    Implementation for the ``murmur3`` hash functions, using the optional dependency `mmh3 <https://github.com/hajimes/mmh3>`_.
"""

from typing import Optional

from multiformats.varint import BytesLike
from .utils import Hashfun, validate_hashfun_args

def _murmur3(version: str, digest_bits: int) -> Hashfun:
    try:
        import mmh3 # pylint: disable = import-outside-toplevel
    except ImportError as e:
        raise ImportError("Module 'mmh3' must be installed to use 'murmur3' hash functions. Consider running 'pip install mmh3'.") from e
    if version == "32":
        def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
            validate_hashfun_args(data, size, 4)
            if not isinstance(data, bytes):
                data = bytes(data)
            d: bytes = mmh3.hash(data, signed=False).to_bytes(4, byteorder="big") # pylint: disable = c-extension-no-member
            return d if size is None else d[:size]
    elif digest_bits == 128: # version == "x64"
        def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
            validate_hashfun_args(data, size, 16)
            if not isinstance(data, bytes):
                data = bytes(data)
            d: bytes = mmh3.hash128(data, signed=False).to_bytes(16, byteorder="big") # pylint: disable = c-extension-no-member
            return d if size is None else d[:size]
    else: # version == "x64"
        def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
            validate_hashfun_args(data, size, 8)
            if not isinstance(data, bytes):
                data = bytes(data)
            d: bytes = mmh3.hash128(data, signed=False).to_bytes(16, byteorder="big") # pylint: disable = c-extension-no-member
            return d[:8] if size is None else d[:size]
    return hashfun

def _jit_register_murmur3(m, register) -> bool: # type: ignore
    if m[1] == "32":
        if register is not None:
            register("murmur3-32", _murmur3("32", 32), 32//8)
        return True
    # version == "x64"
    assert m[2] == "x64"
    digest_bits = int(m[3])
    if register is not None:
        register(f"murmur3-x64-{digest_bits}", _murmur3("x64", digest_bits), digest_bits//8)
    return True



"""
    Implementation for the ``blake2`` and ``blake3`` hash functions, using the optional dependency `blake3 <https://github.com/oconnor663/blake3-py>`_.
"""

import hashlib
from typing import Optional

from multiformats.varint import BytesLike
from .utils import Hashfun, validate_hashfun_args

def _hashlib_blake2(version: str, digest_bits: int) -> Hashfun:
    h = getattr(hashlib, f"blake2{version}")
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, digest_bits//8)
        m: hashlib._Hash = h(digest_size=digest_bits//8) # pylint: disable = no-member
        m.update(data)
        d = m.digest()
        return d if size is None else d[:size]
    return hashfun

def _jit_register_blake2(m, register) -> bool: # type: ignore
    blake2_version, digest_bits = (m[1], int(m[2]))
    if digest_bits not in range(8, 513 if blake2_version == "b" else 257, 8):
        return False
    if register is not None:
        register(f"blake2{blake2_version}-{digest_bits}", _hashlib_blake2(blake2_version, digest_bits), digest_bits//8)
    return True

def _blake3() -> Hashfun:
    try:
        from blake3 import blake3 # type: ignore # pylint: disable = import-outside-toplevel
    except ImportError as e:
        raise ImportError("Module 'blake3' must be installed to use 'blake3' hash function. Consider running 'pip install blake3'.") from e
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, None, size_required=True, name="blake3")
        assert size is not None
        m = blake3() # pylint: disable = not-callable
        m.update(data)
        d: bytes = m.digest(size)
        return d
    return hashfun

def _jit_register_blake3(m, register) -> bool: # type: ignore
    if register is not None:
        register("blake3", _blake3(), None)
    return True



r"""
    Implementation of raw data encodings used by multibase encodings.

    The majority of the encodings is provided by the `bases <https://github.com/hashberg-io/bases>`_ library,
    as instances of its :class:`~bases.encoding.base.BaseEncoding` class. The following custom encodings are also implemented:

    - multibase identity
    - multibase proquints

    Core functionality is provided by the :func:`get` and :func:`exists` functions,
    which can be used to check whether a raw encoding with given name is known, and if so to get the corresponding object:

    >>> from multiformats.multibase import raw_encoding
    >>> raw_encoding.exists("base10")
    True
    >>> raw_encoding.get("base10")
    ZeropadBaseEncoding(StringAlphabet('0123456789'))

    The raw encoding objects have :meth:`CustomEncoding.encode` and
    :meth:`CustomEncoding.decode` methods that can be used to convert between
    bytestrings and strings (not including the multibase code):

    >>> base16 = raw_encoding.get("base16")
    >>> base16.encode(bytes([0xAB, 0xCD]))
    'abcd'
    >>> base16.decode('abcd')
    b'\xab\xcd'

"""

import binascii
from itertools import product
from types import MappingProxyType
from typing import Any, Callable, Dict, List, Tuple, Union
from typing_extensions import Literal
from typing_validation import validate

from bases import (base2, base16, base8, base10, base36, base58btc, base58flickr,
                   base32, base32hex, base32z, base64, base64url,)
from bases.encoding import BaseEncoding

from multiformats.varint import BytesLike
from .err import MultibaseKeyError, MultibaseValueError

RawEncoder = Callable[[BytesLike], str]
""" Type alias for a raw base encoder. """

RawDecoder = Callable[[str], bytes]
""" Type alias for a raw base decoder. """

class CustomEncoding:
    """
        Class for custom raw encodings, implemented by explicitly passing raw encoding and decoding functions.
        The raw encoder and decoder are expected to validate their own arguments.
    """

    _raw_encoder: RawEncoder
    _raw_decoder: RawDecoder

    def __init__(self, raw_encoder: RawEncoder, raw_decoder: RawDecoder):
        # validate(raw_encoder, Callable[[bytes], str]) # TODO: not yet supported by typing-validation
        # validate(raw_decoder, Callable[[str], bytes]) # TODO: not yet supported by typing-validation
        self._raw_encoder = raw_encoder
        self._raw_decoder = raw_decoder

    def encode(self, b: BytesLike) -> str:
        """
            Calls the custom raw encoder.

            :param b: the bytestring to be encoded
            :type b: :obj:`~multiformats.varint.BytesLike`
        """
        raw_encoder: RawEncoder = self._raw_encoder
        return raw_encoder(b)

    def decode(self, s: str) -> bytes:
        """
            Calls the custom raw decoder.

            :param s: the string to be decoded
            :type s: :obj:`str`
        """
        raw_decoder: RawDecoder = self._raw_decoder
        return raw_decoder(s)

    def __repr__(self) -> str:
        _raw_encoder: Callable[[bytes], str] = self._raw_encoder
        _raw_decoder: Callable[[str], bytes] = self._raw_decoder
        return f"CustomEncoding({repr(_raw_encoder)}, {repr(_raw_decoder)})"


RawEncoding = Union[CustomEncoding, BaseEncoding]
_raw_encodings: Dict[str, RawEncoding] = {}

def get(name: str) -> RawEncoding:
    """
        Gets the raw encoding with given name.

        Example usage:

        >>> raw_encoding.get("base16")
        ZeropadBaseEncoding(
            StringAlphabet('0123456789abcdef',
                           case_sensitive=False),
            block_nchars=2)

        :param name: the name for the encoding
        :type name: :obj:`str`

        :raises KeyError: if no such encoding exists

        :rtype: :class:`CustomEncoding` or :class:`~bases.encoding.base.BaseEncoding`
    """
    validate(name, str)
    if name not in _raw_encodings:
        if not _jit_register_encoding(name):
            raise MultibaseKeyError(f"No raw encoding named {repr(name)}.")
    return _raw_encodings[name]


def exists(name: str) -> bool:
    """
        Checks whether a raw encoding with given name exists.

        Example usage:

        >>> raw_encoding.exists("base16")
        True

        :param name: the name for the encoding
        :type name: :obj:`str`

    """
    validate(name, str)
    return name in _raw_encodings or name in _jit_registered_encodings


def register(name: str, enc: RawEncoding, *, overwrite: bool = False) -> None:
    """
        Registers a raw encoding by name.

        Example usage:

        >>> from bases import base45
        >>> raw_encoding.register("base45upper", base45)
        >>> raw_encoding.get("base45upper")
        BlockBaseEncoding(
            StringAlphabet('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:',
                           case_sensitive=False),
            block_size={1: 2, 2: 3}, reverse_blocks=True)

        :param name: the name for the encoding being registered
        :type name: :obj:`str`
        :param enc: the raw encoding being registered
        :type enc: :class:`~bases.encoding.base.BaseEncoding` or :class:`CustomEncoding`
        :param overwrite: whether to overwrite an existing encoding with same name (default :obj:`False`)
        :type overwirte: :obj:`bool`, *optional*

        :raises ValueError: if ``overwrite`` is :obj:`False` and a raw encoding with the same name already exists.
    """
    validate(name, str)
    validate(enc, RawEncoding)
    validate(overwrite, bool)
    if not overwrite and name in _raw_encodings:
        raise MultibaseValueError(f"Raw encoding with name {repr(name)} already exists: {_raw_encodings[name]}")
    _raw_encodings[name] = enc


def unregister(name: str) -> None:
    """
        Unregisters a raw encoding by name.

        Example usage:

        >>> raw_encoding.unregister("base45upper")
        >>> raw_encoding.exists("base45upper")
        False

        :param name: the raw encoding name to unregister
        :type name: :obj:`str`
        :raises KeyError: if no such raw encoding exists
    """
    validate(name, str)
    if name not in _raw_encodings:
        raise MultibaseKeyError(f"Raw encoding with name {repr(name)} does not exist.")
    del _raw_encodings[name]


# register base encodings already instantiated by 'bases' v0.2.1
register("base2", base2)
register("base8", base8)
register("base10", base10)
register("base16upper", base16)
register("base32padupper", base32)
register("base32hexpadupper", base32hex)
register("base32z", base32z)
register("base36upper", base36)
register("base58btc", base58btc)
register("base58flickr", base58flickr)
register("base64pad", base64)
register("base64urlpad", base64url)


def _jit_register_identity_encoding() -> None:
    def identity_raw_encoder(b: BytesLike) -> str:
        """
            Implementation of the raw identity encoder according to the `multibase spec <https://github.com/multiformats/multibase/>`_.
        """
        validate(b, Union[bytes, bytearray, memoryview])
        return str(b, "utf-8")
        # if isinstance(b, (bytes, bytearray)):
        #     return b.decode("utf-8")
        # validate(b, memoryview)
        # return bytes(b).decode("utf-8")
    identity_raw_encoder.__repr__ = lambda: "identity_raw_encoder" # type: ignore
    def identity_raw_decoder(s: str) -> bytes:
        """
            Implementation of the raw identity decoder according to the `multibase spec <https://github.com/multiformats/multibase/>`_.
        """
        validate(s, str)
        return s.encode("utf-8")
    identity_raw_decoder.__repr__ = lambda: "identity_raw_decoder" # type: ignore
    register("identity", CustomEncoding(identity_raw_encoder, identity_raw_decoder))

def _jit_register_proquint_encoding() -> None:
    # pylint: disable = too-many-statements
    _proquint_consonants = "bdfghjklmnprstvz"
    _proquint_consonants_set = frozenset("bdfghjklmnprstvz")
    _proquint_vowels = "aiou"
    _proquint_vowels_set = frozenset("aiou")
    _proquint_consonants_revdir = MappingProxyType({char: idx for idx, char in enumerate(_proquint_consonants)})
    _proquint_vowels_revdir = MappingProxyType({char: idx for idx, char in enumerate(_proquint_vowels)})
    def proquint_raw_encoder(b: BytesLike) -> str:
        """
            Implementation of the proquint encoder according to the `proquint spec <https://arxiv.org/html/0901.4016>`_,
            with additional 'ro-' prefix as prescribed by the `multibase spec <https://github.com/multiformats/multibase/>`_
            and extended to include odd-length bytestrings (adding a final 3-letter block, using two zero pad bits).
        """
        validate(b, BytesLike)
        b = memoryview(b) # makes slicing cheap
        consonants = _proquint_consonants
        vowels = _proquint_vowels
        char_blocks: List[str] = []
        for idx in range(0, len(b), 2):
            byte_block = b[idx: idx+2]
            i = int.from_bytes(byte_block, byteorder="big")
            if len(byte_block) == 2: # ordinary byte pair
                i, c2 = divmod(i, 16) # 4 bits
                i, v1 = divmod(i, 4)  # 2 bits
                i, c1 = divmod(i, 16) # 4 bits
                i, v0 = divmod(i, 4)  # 2 bits
                i, c0 = divmod(i, 16) # 4 bits
                assert i == 0
                char_block = consonants[c0]+vowels[v0]+consonants[c1]+vowels[v1]+consonants[c2]
                char_blocks.append(char_block)
            else: # final byte for odd-length bytestrings
                i <<= 2 # add 2 zero pad bits
                i, c1 = divmod(i, 16) # 4 bits
                i, v0 = divmod(i, 4)  # 2 bits
                i, c0 = divmod(i, 16) # 4 bits
                assert i == 0
                char_block = consonants[c0]+vowels[v0]+consonants[c1]
                char_blocks.append(char_block)
        prefix = "ro-" # follows multibase code "p" to make "pro-", e.g. "pro-lusab-babad"
        return prefix+"-".join(char_blocks)
    proquint_raw_encoder.__repr__ = lambda: "proquint_raw_encoder" # type: ignore
    def proquint_raw_decoder(s: str) -> bytes:
        """
            Implementation of the proquint decoder according to the `proquint spec <https://arxiv.org/html/0901.4016>`_,
            with additional 'ro-' prefix as prescribed by the `multibase spec <https://github.com/multiformats/multibase/>`_
            and extended to include odd-length bytestrings (adding a final 3-letter block, using two zero pad bits).
        """
        # pylint: disable = too-many-branches
        validate(s, str)
        consonants = _proquint_consonants
        vowels = _proquint_vowels
        consonants_set = _proquint_consonants_set
        vowels_set = _proquint_vowels_set
        consonants_revdir = _proquint_consonants_revdir
        vowels_revdir = _proquint_vowels_revdir
        # validate string
        if not s.startswith("ro-"):
            raise binascii.Error("Multibase proquint encoded strings must start with 'ro-'.")
        # remove 'ro-' prefix, return empty bytestring if resultant string is empty
        s = s[3:]
        if len(s) == 0:
            return b""
        # validate length for patterns cvcvc (len 5), cvcvc-...-cvc (len 6k+3) or cvcvc-...-cvcvc (len 6k+5)
        if len(s) % 6 not in (3, 5):
            raise binascii.Error("Proquint encoded string length must give remainder of 3 or 5 when divided by 6.")
        # validate characters and convert encoded string into unsigned integer
        i = 0
        for idx, char in enumerate(s):
            if idx % 6 == 5: # separator
                if char != "-":
                    raise binascii.Error(f"Incorrect char at position {idx}: expected '-', found {repr(char)}.")
            elif idx % 2 == 0: # consonant
                if char not in consonants_set:
                    raise binascii.Error(f"Incorrect char at position {idx}: expected consonant in {repr(consonants)}, "
                                         f"found {repr(char)}.")
                i <<= 4 # make space for 4 bits
                i += consonants_revdir[char] # insert consonant bits
            else: # vowel
                if char not in vowels_set:
                    raise binascii.Error(f"Incorrect char at position {idx}: expected vowel in {repr(vowels)}, "
                                         f"found {repr(char)}.")
                i <<= 2 # make space for 2 bits
                i += vowels_revdir[char] # insert vowel bits
        # set number of bytes to number of quintuplets
        nbytes = 2*((len(s)+1)//6)
        # deal with the case of terminating tripled (odd bytestring length)
        if len(s) % 6 == 3:
            # ensure pad bits are zero
            i, pad_bits = divmod(i, 4)
            if pad_bits != 0:
                raise binascii.Error(f"Expected pad bits to be 00, found {bin(pad_bits)[2:]} instead.")
            # add an extra byte
            nbytes += 1
        # convert unsigned integer to bytes and return
        return i.to_bytes(nbytes, byteorder="big")
    proquint_raw_decoder.__repr__ = lambda: "proquint_raw_decoder" # type: ignore
    register("proquint", CustomEncoding(proquint_raw_encoder, proquint_raw_decoder))

def _jit_register_base_encoding(b: Literal[16, 32, 36, 64],
                       _hex: bool = False,
                       _pad: bool = False,
                       _upper: bool = False,
                       _url: bool = False) -> None:
    if b == 64:
        assert not _hex and not _pad and not _upper
        if _url:
            register("base64url", base64url.nopad())
        else:
            register("base64", base64.nopad())
        return
    assert not _url
    if b in (16, 36):
        assert not _hex and not _pad and not _upper
        if b == 16:
            register("base16", base16.lower())
        else:
            register("base36", base36.lower())
        return
    assert b == 32 and (not _pad or not _upper)
    base = base32hex if _hex else base32
    if not _pad:
        base = base.nopad()
    if not _upper:
        base = base.lower()
    key = f"base32{'hex' if _hex else ''}{'pad' if _pad else ''}{'upper' if _upper else ''}"
    register(key, base)

_jit_registered_encodings: Dict[str, Tuple[Callable[..., Any], Any]] = {
    "identity": (_jit_register_identity_encoding, tuple()),
    "proquint": (_jit_register_proquint_encoding, tuple()),
    **{
        f"base64{'url' if _url else ''}": (
            _jit_register_base_encoding,
            (64, False, False, False, _url)
        )
        for _url in (False, True)
    },
    **{
        f"base{b}": (
            _jit_register_base_encoding,
            (b,)
        )
        for b in (16, 36)
    },
    **{
        f"base32{'hex' if _hex else ''}{'pad' if _pad else ''}{'upper' if _upper else ''}": (
            _jit_register_base_encoding,
            (32, _hex, _pad, _upper)
        )
        for _hex, _pad, _upper in product((False, True), repeat=3)
    }
}

def _jit_register_encoding(name: str) -> bool:
    if name not in _jit_registered_encodings:
        return False
    f, args = _jit_registered_encodings[name]
    f(*args)
    return True



"""
    Errors for the :mod:`~multiformats.multiaddr` module.
"""

import builtins

class MultiaddrKeyError(builtins.KeyError): # pylint: disable = redefined-builtin
    """ Class for :mod:`~multiformats.multiaddr` key errors. """


class MultiaddrValueError(builtins.ValueError): # pylint: disable = redefined-builtin
    """ Class for :mod:`~multiformats.multiaddr` value errors. """



"""
    Implementation of the `CID spec <https://github.com/multiformats/cid>`_.

    This module differs from other modules of :mod:`~multiformats`, in that the functionality is completely
    encapsulated by a single class :class:`CID`, which is imported from top level instead
    of the module itself:

    >>> from multiformats import CID
"""

from typing import Any, cast, FrozenSet, Tuple, Type, TypeVar, Union
from typing_extensions import Literal, Final
from typing_validation import validate

from bases import base58btc
from multiformats import varint, multicodec, multibase, multihash

from multiformats.multicodec import Multicodec
from multiformats.multibase import Multibase
from multiformats.multihash import Multihash, _validate_raw_digest_size
from multiformats.varint import BytesLike, byteslike

_CIDSubclass = TypeVar("_CIDSubclass", bound="CID")

CIDVersion = Literal[0, 1]
CIDVersionNumbers: Final[FrozenSet[int]] = frozenset({0, 1})

def _binary_cid_from_str(cid: str) -> Tuple[bytes, Multibase]:
    if len(cid) == 46 and cid.startswith("Qm"):
        # CIDv0 to be decoded as base58btc
        return base58btc.decode(cid), multibase.get("base58btc")
    mb, b = multibase.decode_raw(cid)
    if b[0] ==  0x12:
        # CIDv0 may not be multibase encoded (0x12 is the first byte of sha2-256 multihashes)
        # CIDv18 (first byte 18=0x12) will be skipped to prevent ambiguity
        raise ValueError("CIDv0 may not be multibase encoded (found multibase encoded bytes starting with 0x12).")
    return b, mb

def _CID_validate_multibase(base: Union[str, Multibase]) -> Multibase:
    if isinstance(base, str):
        base = multibase.get(base)
    else:
        multibase.validate_multibase(base)
    return base

def _CID_validate_multicodec(codec: Union[str, int, Multicodec]) -> Multicodec:
    if isinstance(codec, str):
        codec = multicodec.get(codec)
    elif isinstance(codec, int):
        codec = multicodec.get(code=codec)
    else:
        multicodec.validate_multicodec(codec)
    return codec

def _CID_validate_multihash(hashfun: Union[str, int, Multihash]) -> Multihash:
    if isinstance(hashfun, str):
        hashfun = multihash.get(hashfun)
    elif isinstance(hashfun, int):
        hashfun = multihash.get(code=hashfun)
    else:
        pass
    return hashfun

def _CID_validate_raw_digest(raw_digest: Union[str, BytesLike], hashfun: Multihash) -> bytes:
    if isinstance(raw_digest, str):
        raw_digest = bytes.fromhex(raw_digest)
    else:
        validate(raw_digest, BytesLike)
        if not isinstance(raw_digest, bytes):
            raw_digest = bytes(raw_digest)
    _, max_digest_size = hashfun.implementation
    _validate_raw_digest_size(hashfun.name, raw_digest, max_digest_size)
    return raw_digest

def _CID_validate_multihash_digest(digest: Union[str, BytesLike]) -> Tuple[Multihash, bytes]:
    if isinstance(digest, str):
        digest = bytes.fromhex(digest)
    raw_digest: BytesLike
    code, raw_digest = multihash.unwrap_raw(digest)
    hashfun = _CID_validate_multihash(code)
    raw_digest = _CID_validate_raw_digest(raw_digest, hashfun)
    return hashfun, raw_digest

def _CID_validate_version(version: int, base: Multibase, codec: Multicodec, hashfun: Multihash) -> int:
    if version in (2, 3):
        raise ValueError("CID versions 2 and 3 are reserved for future use.")
    if version not in (0, 1):
        raise ValueError(f"CID version {version} is not allowed.")
    if version == 0:
        if base.name != 'base58btc':
            raise ValueError(f"CIDv0 multibase must be 'base58btc', found {repr(base.name)} instead.")
        if codec.name != "dag-pb":
            raise ValueError(f"CIDv0 multicodec must be 'dag-pb', found {repr(codec.name)} instead.")
        if hashfun.name != "sha2-256":
            raise ValueError(f"CIDv0 multihash must be 'sha2-256', found {repr(hashfun.name)} instead.")
    return version


class CID:
    """

        Container class for `Content IDentifiers <https://github.com/multiformats/cid>`_.

        CIDs can be explicitly instantiated by passing multibase, CID version, multicodec and multihash digest to the constructor:

        >>> cid = CID("base58btc", 1, "raw",
        ... "12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95")
        >>> str(cid)
        'zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA'

        Alternatively, a pair of multihash codec and raw hash digest can be passed in lieu of the multihash digest:

        >>> raw_digest = bytes.fromhex(
        ... "6e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95")
        >>> cid = CID("base58btc", 1, "raw", ("sha2-256", raw_digest))
        >>> str(cid)
        'zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA'

        The multihash digest and raw digest values can be passed either as :obj:`bytes`-like objects or as the corresponding hex strings:

        >>> isinstance(raw_digest, bytes)
        True
        >>> raw_digest.hex()
        '6e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95'

        Note: the hex strings are not multibase encoded.

        Calling :obj:`bytes` on an instance of this class returns its binary representation, as a :obj:`bytes` object:

        >>> cid = CID("base58btc", 1, "raw",
        ... "12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95")
        >>> raw_digest.hex()
                '6e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95'
        >>> bytes(cid).hex()
        '015512206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95'
        #^^   0x01 = CIDv1
        #  ^^ 0x55 = 'raw' codec
        >>> bytes(cid)

        :param base: default multibase to use when encoding this CID
        :type base: :obj:`str` or :class:`~multiformats.multibase.Multibase`
        :param version: the CID version
        :type version: 0 or 1
        :param codec: the content multicodec
        :type codec: :obj:`str`, :obj:`int` or :class:`~multiformats.multicodec.Multicodec`
        :param digest: the content multihash digest, or a pair of multihash codec and raw content digest
        :type digest: see below

        The ``digest`` parameter can be specified in the following ways:

        - as a :obj:`str`, in which case it is treated as a hex-string and converted to :obj:`bytes` using :obj:`bytes.fromhex`
        - as a :obj:`~multiformats.varint.BytesLike`, in which case it is converted to :obj:`bytes` directly
        - as a pair ``(multihash_codec, raw_digest)`` of a multihash and raw hash digest, which are used to produce a multihash digest
          via the :meth:`~multiformats.multihash.Multihash.wrap` metho

        If ``digest`` is specified by a pair, the ``multihash_codec`` value can be specified in the following ways:

        - by multihash multicodec name, as a :obj:`str`
        - by multihash multicodec code, as a :obj:`int`
        - as a :class:`~multiformats.multihash.Multihash` object

        If ``digest`` is specified by a pair, the ``raw_digest`` value can be specified in the following ways:

        - as a :obj:`str`, in which case it is treated as a hex-string and converted to :obj:`bytes` using :obj:`bytes.fromhex`
        - as a :obj:`~multiformats.varint.BytesLike`, in which case it is converted to :obj:`bytes` directly


        :raises ValueError: if the CID version is unsupported
        :raises ValueError: if version is 0 but base is not 'base58btc' or codec is not 'dag-pb'
        :raises KeyError: if the multibase, multicodec or multihash are unknown

    """

    _base: Multibase
    _version: CIDVersion
    _codec: Multicodec
    _hashfun: Multihash
    _digest: bytes

    __slots__ = ("__weakref__", "_base", "_version", "_codec", "_hashfun", "_digest")

    def __new__(cls: Type[_CIDSubclass],
                base: Union[str, Multibase],
                version: int,
                codec: Union[str, int, Multicodec],
                digest: Union[str, BytesLike, Tuple[Union[str, int, Multihash], Union[str, BytesLike]]],
                ) -> _CIDSubclass:
        # pylint: disable = too-many-arguments
        base = _CID_validate_multibase(base)
        codec = _CID_validate_multicodec(codec)
        raw_digest: Union[str, bytes]
        hashfun: Union[str, int, Multihash]
        if isinstance(digest, (str,)+byteslike):
            hashfun, raw_digest = _CID_validate_multihash_digest(digest)
        else:
            validate(digest, Tuple[Union[str, int, Multihash], Union[str, BytesLike]])
            hashfun, raw_digest = digest
            hashfun = _CID_validate_multihash(hashfun)
            raw_digest = _CID_validate_raw_digest(raw_digest, hashfun)
        version = _CID_validate_version(version, base, codec, hashfun)
        if isinstance(digest, bytes):
            return CID._new_instance(cls, base, version, codec, hashfun, digest)
        return CID._new_instance(cls, base, version, codec, hashfun, (hashfun, raw_digest))

    @staticmethod
    def _new_instance(CID_subclass: Type[_CIDSubclass],
                      base: Multibase,
                      version: int,
                      codec: Multicodec,
                      hashfun: Multihash,
                      digest: Union[bytes, Tuple[Multihash, bytes]],
                     ) -> _CIDSubclass:
        # pylint: disable = too-many-arguments
        instance: _CIDSubclass = object.__new__(CID_subclass)
        instance._base = base
        assert version in (0, 1)
        instance._version = cast(Literal[0, 1], version)
        instance._codec = codec
        instance._hashfun = hashfun
        if isinstance(digest, bytes):
            instance._digest = digest
        elif isinstance(digest, byteslike):
            instance._digest = bytes(digest)
        else:
            _hashfun, raw_digest = digest
            if not isinstance(raw_digest, bytes):
                raw_digest = bytes(raw_digest)
            assert _hashfun == hashfun, "You passed different multihashes to a _new_instance call with digest as a pair."
            instance._digest = hashfun.wrap(raw_digest)
        return instance

    @property
    def version(self) -> CIDVersion:
        """
            CID version.

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid.version
            1

        """
        return self._version

    @property
    def base(self) -> Multibase:
        """
            Multibase used to encode the CID:

            - if a CIDv1 was decoded from a multibase-encoded string, the encoding multibase is used
            - if a CIDv1 was decoded from a bytestring, the 'base58btc' multibase is used
            - for a CIDv0, 'base58btc' is always used

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid.base
            Multibase(name='base58btc', code='z',
                      status='default', description='base58 bitcoin')

        """
        return self._base

    @property
    def codec(self) -> Multicodec:
        """
            Codec that the multihash digest refers to.

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid.codec
            Multicodec(name='raw', tag='ipld', code='0x55',
                       status='permanent', description='raw binary')

        """
        return self._codec

    @property
    def hashfun(self) -> Multihash:
        """
            Multihash used to produce the multihash digest.

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid.hashfun
            Multicodec(name='sha2-256', tag='multihash', code='0x12',
                       status='permanent', description='')

        """
        return self._hashfun

    @property
    def digest(self) -> bytes:
        """
            Multihash digest.

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid.digest.hex()
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95'

        """
        return self._digest

    @property
    def raw_digest(self) -> bytes:
        """
            Raw hash digest, decoded from the multihash digest.

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid.raw_digest.hex()
            '6e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95'

        """
        return multihash.unwrap(self._digest)

    @property
    def human_readable(self) -> str:
        """
            Human-readable representation of the CID.

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid.human_readable
            'base58btc - cidv1 - raw - (sha2-256 : 256 : 6E6FF7950A36187A801613426E858DCE686CD7D7E3C0FC42EE0330072D245C95)'

        """
        raw_digest = self.raw_digest
        hashfun_str = f"({self.hashfun.name} : {len(raw_digest)*8} : {raw_digest.hex().upper()})"
        return f"{self.base.name} - cidv{self.version} - {self.codec.name} - {hashfun_str}"

    def encode(self, base: Union[None, str, Multibase] = None) -> str:
        """
            Encodes the CID using a given multibase. If :obj:`None` is given,
            the CID's own multibase is used by default.

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid.encode() # default: cid.base
            'zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA'
            >>> cid.encode("base32")
            'bafkreidon73zkcrwdb5iafqtijxildoonbwnpv7dyd6ef3qdgads2jc4su'

            :param base: the multibase to be used for encoding
            :type base: :obj:`None`, :obj:`str` or :class:`~multiformats.multibase.Multibase`, *optional*

            :raises KeyError: see :meth:`multiformats.multibase.Multibase.encode`

        """
        if self.version == 0:
            if base is not None:
                raise ValueError("CIDv0 cannot be multibase-encoded, please set multibase=None.")
            return base58btc.encode(bytes(self))
        if base is None or base == self.base:
            base = self.base # use CID's own multibase as default
        else:
            if isinstance(base, str):
                base = multibase.get(base)
            else:
                multibase.validate_multibase(base)
        return base.encode(bytes(self))

    def set(self, *,
            base: Union[None, str, Multibase] = None,
            version: Union[None, int] = None,
            codec: Union[None, str, int, Multicodec] = None
           ) -> "CID":
        """
            Returns a new CID obtained by setting new values for one or more of:
            ``base``, ``version``, or ``codec``.

            Example usage:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid
            CID('base58btc', 1, 'raw',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')
            >>> cid.set(base="base32")
            CID('base32', 1, 'raw',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')
            >>> cid.set(codec="dag-cbor")
            CID('base58btc', 1, 'dag-cbor',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')
            >>> cid.set(version=0, codec="dag-pb")
            CID('base58btc', 0, 'dag-pb',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')
            >>> cid
            CID('base58btc', 1, 'raw',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')
            # Note: 'CID.set' returns new instances,
            #       the original 'cid' instance is unchanged

            If setting ``version`` to 0, ``base`` must be 'base58btc' and ``codec`` must be 'dag-pb'.

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> cid = CID.decode(s)
            >>> cid
            CID('base58btc', 1, 'raw',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')
            >>> cid.set(version=0, codec="dag-pb")
            CID('base58btc', 0, 'dag-pb',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')
            >>> cid.set(version=0)
            ValueError: CIDv0 multicodec must be 'dag-pb', found 'raw' instead.
            >>> cid.set(version=0, codec="dag-pb", base="base32")
            ValueError: CIDv0 multibase must be 'base58btc', found 'base32' instead

            :param base: the new CID multibase, or :obj:`None` if multibase unchanged
            :type base: :obj:`None`, :obj:`str` or :class:`~multiformats.multibase.Multibase`, *optional*
            :param version: the new CID version, or :obj:`None` if version unchanged
            :type version: :obj:`None`, 0 or 1, *optional*
            :param codec: the new content multicodec, or :obj:`None` if multicodec unchanged
            :type codec: :obj:`None`, :obj:`str` or :class:`~multiformats.multicodec.Multicodec`, *optional*

            :raises KeyError: if the multibase or multicodec are unknown

        """
        hashfun = self.hashfun
        digest = self.digest
        if base is not None and base not in (self.base, self.base.name):
            base = _CID_validate_multibase(base)
        else:
            base = self.base
        if codec is not None and codec not in (self.codec, self.codec.name, self.codec.code):
            codec = _CID_validate_multicodec(codec)
        else:
            codec = self.codec
        if version is not None and version != self.version:
            _CID_validate_version(version, base, codec, hashfun)
        else:
            version = self.version
        return CID._new_instance(CID, base, version, codec, hashfun, digest)

    def __bytes__(self) -> bytes:
        if self.version == 0:
            return self.digest
        return varint.encode(self.version)+varint.encode(self.codec.code)+self.digest

    def __str__(self) -> str:
        return self.encode()

    def __repr__(self) -> str:
        mb = self.base.name
        v = self.version
        mc = self.codec.name
        d = self.digest
        return f"CID({repr(mb)}, {v}, {repr(mc)}, {repr(d.hex())})"

    @property
    def _as_tuple(self) -> Tuple[Type["CID"], int, Multicodec, bytes]:
        return (CID, self.version, self.codec, self.digest)

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, CID):
            return NotImplemented
        return self._as_tuple == other._as_tuple

    @staticmethod
    def decode(cid: Union[str, BytesLike]) -> "CID":
        """
            Decodes a CID from a bytestring or a hex string (which will be converted to :obj:`bytes`
            using :obj:`bytes.fromhex`). Note: the hex string is not multibase encoded.

            Example usage for CIDv1 multibase-encoded string:

            >>> s = "zb2rhe5P4gXftAwvA4eXQ5HJwsER2owDyS9sKaQRRVQPn93bA"
            >>> CID.decode(s)
            CID('base58btc', 1, 'raw',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')

            Example usage for CIDv1 bytestring (multibase always set to 'base58btc'):

            >>> b = bytes.fromhex(
            ... "015512206e6ff7950a36187a801613426e85"
            ... "8dce686cd7d7e3c0fc42ee0330072d245c95")
            >>> CID.decode(b)
            CID('base58btc', 1, 'raw',
            '12206e6ff7950a36187a801613426e858dce686cd7d7e3c0fc42ee0330072d245c95')

            Example usage for CIDv0 base58-encoded string:

            >>> s = "QmbWqxBEKC3P8tqsKc98xmWNzrzDtRLMiMPL8wBuTGsMnR"
            >>> CID.decode(s)
            CID('base58btc', 0, 'dag-pb',
            '1220c3c4733ec8affd06cf9e9ff50ffc6bcd2ec85a6170004bb709669c31de94391a')

            Example usage for CIDv0 bytestring (multibase always set to 'base58btc'):

            >>> b = bytes.fromhex(
            ... "1220c3c4733ec8affd06cf9e9ff50ffc6b"
            ... "cd2ec85a6170004bb709669c31de94391a")
            >>> CID.decode(b)
            CID('base58btc', 0, 'dag-pb',
            '1220c3c4733ec8affd06cf9e9ff50ffc6bcd2ec85a6170004bb709669c31de94391a')

            :param cid: the CID bytes or multibase-encoded string
            :type cid: :obj:`str` or :obj:`~multiformats.varint.BytesLike`

            :raises ValueError: if the CID is malformed or the CID version is unsupported
            :raises KeyError: if the multibase, multicodec or multihash are unknown

        """
        if isinstance(cid, str):
            cid, mb = _binary_cid_from_str(cid)
        else:
            mb = multibase.get("base58btc")
        validate(cid, BytesLike)
        cid = memoryview(cid)
        # if len(cid) == 34 and cid.startswith(b"\x12\x20"):
        if len(cid) == 34 and cid[0] == 0x12 and cid[1] == 0x20:
            v = 0 # CID version
            mc_code = 0x70 # multicodec.get("dag-pb")
            digest = cid  # multihash digest is what's left
        else:
            v, _, cid = varint.decode_raw(cid) # CID version
            if v == 0:
                raise ValueError("CIDv0 is malformed.")
            if v in (2, 3):
                raise ValueError("CID versions 2 and 3 are reserved for future use.")
            if v != 1:
                raise ValueError(f"CIDv{v} is currently not supported.")
            mc_code, _, cid = multicodec.unwrap_raw(cid) # multicodec
            digest = cid # multihash digest is what's left
        mc = multicodec.get(code=mc_code)
        mh_code, _ = multihash.unwrap_raw(digest)
        mh = multihash.get(code=mh_code)
        return CID._new_instance(CID, mb, v, mc, mh, digest)

    @staticmethod
    def peer_id(pk_bytes: Union[str, BytesLike]) -> "CID":
        """
            Wraps the raw hash of a public key into a `PeerID <https://docs.libp2p.io/concepts/peer-id/>`_, as a CIDv1.

            The ``pk_bytes`` argument should be the binary public key, encoded according to the
            `PeerID spec <https://github.com/libp2p/specs/blob/master/peer-ids/peer-ids.md>`_.
            This can be passed as a bytestring or as a hex string (which will be converted to :obj:`bytes` using :obj:`bytes.fromhex`).
            Note: the hex string is not multibase encoded.

            Example usage with Ed25519 public key:

            >>> pk_bytes = bytes.fromhex(
            ... "1498b5467a63dffa2dc9d9e069caf075d16fc33fdd4c3b01bfadae6433767d93")
            ... # a 32-byte Ed25519 public key
            >>> peer_id = CID.peer_id(pk_bytes)
            >>> peer_id
            CID('base32', 1, 'libp2p-key',
            '00201498b5467a63dffa2dc9d9e069caf075d16fc33fdd4c3b01bfadae6433767d93')
            #^^   0x00 = 'identity' multihash used (public key length <= 42)
            #  ^^ 0x20 = 32-bytes of raw hash digestlength
            >>> str(peer_id)
            'bafzaaiautc2um6td375c3soz4bu4v4dv2fx4gp65jq5qdp5nvzsdg5t5sm'

            Snippet showing how to obtain the `Ed25519 <https://cryptography.io/en/latest/hazmat/primitives/asymmetric/ed25519/>`_
            public key bytestring using the `cryptography <https://github.com/pyca/cryptography>`_ library:

            >>> from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
            >>> from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
            >>> private_key = Ed25519PrivateKey.generate()
            >>> public_key = private_key.public_key()
            >>> pk_bytes = public_key.public_bytes(
            ...     encoding=Encoding.Raw,
            ...     format=PublicFormat.Raw
            ... )
            >>> pk_bytes.hex()
            "1498b5467a63dffa2dc9d9e069caf075d16fc33fdd4c3b01bfadae6433767d93"

            Example usage with DER-encoded RSA public key:

            >>> pk_bytes = bytes.fromhex(
            ... "30820122300d06092a864886f70d01010105000382010f003082010a02820101"
            ... "009a56a5c11e2705d0bfe0cd1fa66d5e519095cc741b62ed99ddf129c32e046e"
            ... "5ba3958bb8a068b05a95a6a0623cc3c889b1581793cd84a34cc2307e0dd74c70"
            ... "b4f230c74e5063ecd8e906d372be4eba13f47d04427a717ac78cb12b4b9c2ab5"
            ... "591f36f98021a70f84d782c36c51819054228ff35a45efa3f82b27849ec89036"
            ... "26b4a4c4b40f9f74b79caf55253687124c79cb10cd3bc73f0c44fbd341e5417d"
            ... "2e85e900d22849d2bc85ca6bf037f1f5b4f9759b4b6942fccdf1140b30ea7557"
            ... "87deb5c373c5953c14d64b523959a76a32a599903974a98cf38d4aaac7e359f8"
            ... "6b00a91dcf424bf794592139e7097d7e65889259227c07155770276b6eda4cec"
            ... "370203010001")
            ... # a 294-byte RSA public key
            >>> peer_id = CID.peer_id(pk_bytes)
            >>> peer_id
            CID('base32', 1, 'libp2p-key',
            '1220c1a6513ffb14f202f75453c49666a5b9d7ed9a1a068891daf824d477573f829f')
            #^^   0x12 = 'sha2-256' multihash used (public key length > 42)
            #  ^^ 0x20 = 32-bytes of raw hash digest length
            >>> str(peer_id)
            'bafzbeigbuzit76yu6ibpovctyslgnjnz27wzugqgrci5v6be2r3vop4ct4'

            Snippet showing how to obtain the `RSA <https://cryptography.io/en/latest/hazmat/primitives/asymmetric/rsa/>`_
            public key bytestring using the `cryptography <https://github.com/pyca/cryptography>`_ library:

            >>> from cryptography.hazmat.primitives.asymmetric import rsa
            >>> from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
            >>> private_key = rsa.generate_private_key(
            ...     public_exponent=65537,
            ...     key_size=2048,
            ... )
            >>> public_key = private_key.public_key()
            >>> pk_bytes = public_key.public_bytes(
            ...     encoding=Encoding.DER,
            ...     format=PublicFormat.SubjectPublicKeyInfo
            ... )
            >>> pk_bytes.hex()
            "30820122300d06092a864886f70d01010105000382010f003082010a02820101"
            "009a56a5c11e2705d0bfe0cd1fa66d5e519095cc741b62ed99ddf129c32e046e"
            "5ba3958bb8a068b05a95a6a0623cc3c889b1581793cd84a34cc2307e0dd74c70"
            "b4f230c74e5063ecd8e906d372be4eba13f47d04427a717ac78cb12b4b9c2ab5"
            "591f36f98021a70f84d782c36c51819054228ff35a45efa3f82b27849ec89036"
            "26b4a4c4b40f9f74b79caf55253687124c79cb10cd3bc73f0c44fbd341e5417d"
            "2e85e900d22849d2bc85ca6bf037f1f5b4f9759b4b6942fccdf1140b30ea7557"
            "87deb5c373c5953c14d64b523959a76a32a599903974a98cf38d4aaac7e359f8"
            "6b00a91dcf424bf794592139e7097d7e65889259227c07155770276b6eda4cec"
            "370203010001"

            :param pk_bytes: the public key bytes
            :type pk_bytes: :obj:`str` or :obj:`~multiformats.varint.BytesLike`

            :raises ValueError: if ``pk_bytes`` is passed as a string and is not the hex-string of some bytes

            """
        if isinstance(pk_bytes, str):
            pk_bytes = bytes.fromhex(pk_bytes)
        else:
            validate(pk_bytes, BytesLike)
        if len(pk_bytes) <= 42:
            mh = multihash.get("identity")
            digest = multihash.digest(pk_bytes, mh)
        else:
            mh = multihash.get("sha2-256")
            digest = multihash.digest(pk_bytes, mh)
        mc = multicodec.get(code=0x72) # multicodec.get("libp2p-key")
        mb = multibase.get("base32")
        return CID._new_instance(CID, mb, 1, mc, mh, digest)



"""
    Implementation of multiformat protocols, according to the `Multiformats <https://multiformats.io/>`_ specifications.

    Suggested usage:

    >>> from multiformats import *

    The above will import the following names:

    .. code-block:: python

        varint, multicodec, multibase, multihash, multiaddr, CID

    The first five are modules implementing homonymous specifications,
    while :class:`~multiformats.cid.CID` is a class for Content IDentifiers.
"""

__version__ = "0.2.1"

from . import varint
from . import multicodec
from . import multibase
from . import multihash
from .cid import CID
from . import multiaddr

__all__ = [
    "varint",
    "multicodec",
    "multibase",
    "multihash",
    "multiaddr",
    "CID",
]



"""
    Utilities for hash function implementation.
"""

from typing import Optional
from typing_extensions import Literal, Protocol, runtime_checkable
from typing_validation import validate
from multiformats.varint import BytesLike
from multiformats.multihash.err import MultihashValueError

@runtime_checkable
class Hashfun(Protocol):
    """
        Protocol for raw hash functions.

        .. code-block:: python

            @runtime_checkable
            class Hashfun(Protocol):
                def __call__(self, data: BytesLike, size: Optional[int] = None) -> bytes:
                    ...

    """

    def __call__(self, data: BytesLike, size: Optional[int] = None) -> bytes:
        ...

def validate_hashfun_args(data: BytesLike,
                          size: Optional[int],
                          max_digest_size: Optional[int],
                          *,
                          size_required: bool = False,
                          name: str = "") -> None:
    """
        Utility function to validate the arguments passed to hash functions.
    """
    validate(data, BytesLike)
    validate(size, Optional[int])
    if size is not None and size < 0:
        raise MultihashValueError("If specified, digest size must be non-negative integer.")
    if size is not None and max_digest_size is not None and size > max_digest_size:
        raise MultihashValueError("If specified, digest size must not exceed maximum digest size for hash function.")
    if size_required and size is None:
        raise MultihashValueError(f"Digest size is mandatory for hash function{' '+name if name else ''}.")

def repeat_hashfun(hashfun: Hashfun,
                   repeat: int = 1,
                   truncate: Literal["end", "always"] = "end") -> Hashfun:
    """
        Utility function for repeated hashing.
    """
    validate(hashfun, Hashfun)
    validate(repeat, int)
    validate(truncate, Literal["end", "always"])
    if repeat <= 0:
        raise MultihashValueError("Argument 'repeat' must be positive integer.")
    if repeat == 1:
        return hashfun
    if truncate == "always":
        def repeated_hashfun(data: BytesLike, size: Optional[int]=None) -> bytes:
            for _ in range(repeat):
                data = hashfun(data, size)
            return data
    else:
        def repeated_hashfun(data: BytesLike, size: Optional[int]=None) -> bytes:
            for _ in range(repeat-1):
                data = hashfun(data, None)
            data = hashfun(data, size)
            return data
    return repeated_hashfun



"""
    Implementation for the ``skein`` hash functions, using the optional dependency `pyskein <https://pythonhosted.org/pyskein/>`_.
"""

import hashlib
from typing import Optional

from multiformats.varint import BytesLike
from .utils import Hashfun, validate_hashfun_args

def _skein(version: int, digest_bits: int) -> Hashfun:
    try:
        import skein # type: ignore # pylint: disable = import-outside-toplevel
    except ImportError as e:
        raise ImportError("Module 'skein' must be installed to use 'skein' hash functions. Consider running 'pip install pyskein'.") from e
    h = getattr(skein, f"skein{version}")
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, digest_bits//8)
        m: hashlib._Hash = h(digest_bits=digest_bits) # pylint: disable = no-member
        m.update(data)
        d = m.digest()
        return d if size is None else d[:size]
    return hashfun

def _jit_register_skein(m, register) -> bool: # type: ignore
    skein_version, digest_bits = (int(m[1]), int(m[2]))
    if digest_bits not in range(8, skein_version+1, 8):
        return False
    if register is not None:
        register(f"skein{skein_version}-{digest_bits}", _skein(skein_version, digest_bits), digest_bits//8)
    return True



"""
    Implementation of the `multiaddr spec <https://github.com/multiformats/multiaddr>`_.

    Suggested usage:

    >>> from multiformats import multiaddr
"""

from itertools import islice, chain
from typing import Any, cast, ClassVar, Dict, Iterator, List, Optional, overload, Sequence, Tuple, Type, Union
from weakref import WeakValueDictionary
import sys
from typing_validation import validate

from multiformats import varint, multicodec
from multiformats.multicodec import Multicodec
from multiformats.varint import BytesLike, byteslike

from . import raw
from .err import MultiaddrKeyError, MultiaddrValueError
from .raw import RawEncoder, RawDecoder, ProtoImpl, _validate_size

class Proto:
    """
        Container class for a single protocol segment of a `multiaddr <https://multiformats.io/multiaddr/>`_.

        >>> ip4 = Proto("ip4")
        >>> ip4
        Proto("ip4")
        >>> str(ip4)
        '/ip4'

        For protocols that don't require an address value, bytes are computed as the varint encoding of protocol code:

        >>> quic = Proto('quic')
        >>> quic.code
        460
        >>> varint.encode(quic.code).hex()
        'cc03'
        >>> bytes(quic).hex()
        'cc03'

        :param codec: the multicodec for this protocol (by name, code, or as object)
        :type codec: :obj:`str`, :obj:`int` or :class:`~multiformats.multicodec.Multicodec`
    """

    # WeakValueDictionary[str, "Proto"]
    _cache: ClassVar[WeakValueDictionary] = WeakValueDictionary() # type: ignore

    _codec: Multicodec
    _implementation: ProtoImpl

    __slots__ = ("__weakref__", "_codec", "_implementation")

    def __new__(cls, codec: Union[str, int, Multicodec]) -> "Proto":
        # check that the codec exists:
        if isinstance(codec, str):
            codec = multicodec.get(codec)
        elif isinstance(codec, int):
            codec = multicodec.get(code=codec)
        else:
            validate(codec, Multicodec)
            existing_codec = multicodec.get(codec.name)
            if existing_codec != codec:
                raise MultiaddrValueError(f"Multicodec named {repr(codec.name)} exists, but is not the one given.")
            codec = existing_codec
        # check that the codec is a multiaddr multicodec:
        if codec.tag != "multiaddr":
            raise MultiaddrValueError(f"Multicodec named {repr(codec.name)} exists, but is not a multiaddr.")
        implementation: ProtoImpl = raw.get(codec.name)
        _cache = Proto._cache
        if codec.name in _cache:
            # if a proto instance with this name is already registered
            instance: Proto = _cache[codec.name]
            if instance._codec == codec and instance._implementation == implementation:
                # nothing changed, can use the existing instance
                return instance
            # otherwise remove the existing instance
            del _cache[codec.name]
        # create a fresh instance, register it and return it
        instance = super().__new__(cls)
        instance._codec = codec
        instance._implementation = implementation
        _cache[codec.name] = instance
        return instance

    @property
    def name(self) -> str:
        """
            Protocol name.

            Example usage:

            >>> ip4.name
            'ip4'

        """
        return self.codec.name

    @property
    def code(self) -> int:
        """
            Protocol code.

            Example usage:

            >>> ip4.code
            4
            # 4 = 0x04

        """
        return self.codec.code

    @property
    def codec(self) -> Multicodec:
        """
            The multicodec for this protocol.

            Example usage:

            >>> ip4.codec
            Multicodec(name='ip4', tag='multiaddr', code='0x04',
                       status='permanent', description='')

        """
        return self._codec

    @property
    def implementation(self) -> ProtoImpl:
        """
            The implementation for this protocol, as a triple of
            raw encoder, raw decoder and address size.

            Example usage:

            >>> ip4.implementation
            (
             <function ip4_encoder at 0x000002B4C9956310>,
             <function ip4_decoder at 0x000002B4C99563A0>,
             4
            )

            :rtype: :obj:`~multiformats.multiaddr.raw.ProtoImpl`
        """
        return self._implementation

    @property
    def raw_encoder(self) -> Optional[RawEncoder]:
        """
            The raw encoder for this protocol.

            Example usage:

            >>> ip4.raw_encoder
            <function ip4_encoder at 0x000002B4C9956310>

            :rtype: :obj:`~multiformats.multiaddr.raw.RawEncoder` or :obj:`None`
        """
        return self.implementation[0]

    @property
    def raw_decoder(self) -> Optional[RawDecoder]:
        """
            The raw decoder for this protocol.

            Example usage:

            >>> ip4.raw_decoder
            <function ip4_decoder at 0x000002B4C99563A0>

            :rtype: :obj:`~multiformats.multiaddr.raw.RawDecoder` or :obj:`None`
        """
        return self.implementation[1]

    @property
    def addr_size(self) -> Optional[int]:
        """
            The address size (in bytes) for this protocol:

            - for protocols with no address, ``addr_size`` is 0
            - for protocols with addresses of variable binary size, ``addr_size`` is :obj:`None`
            - for all other protocols, ``addr_size`` is a positive :obj:`int`

            Example usage:

            >>> ip4.addr_size
            4

        """
        return self.implementation[2]

    @property
    def admits_addr(self) -> bool:
        """
            Whether this protocol admits an address.

            >>> ip4.admits_addr
            True

        """
        return self.addr_size != 0

    def is_addr_valid(self, addr_value: Union[str, BytesLike]) -> bool:
        """
            Validates an address value.

            Example usage:

            >>> ip4.is_addr_valid("192.168.1.1")
            True
            >>> ip4.is_addr_valid(bytes([192, 168, 1, 1]))
            True

            The same result can be obtained with container syntax:

            >>> "192.168.1.1" in ip4
            True
            >>> bytes([192, 168, 1, 1]) in ip4
            True

        """
        try:
            self.validate(addr_value)
            return True
        except MultiaddrValueError:
            return False

    def validate(self, addr_value: Union[str, BytesLike]) -> Tuple[str, bytes]:
        """
            If successful, returns a pair of the string and bytes representations of the address value.

            Example usage:

            >>> ip4.validate("192.168.1.1")
            ('192.168.1.1', b'\\xc0\\xa8\\x01\\x01')
            >>> ip4.validate("192.168")
            MultiaddrValueError: Expected 4 octets in '192.168'

            :raises ValueError: if ``not self.is_valid(addr_value)``
        """
        raw_encoder, raw_decoder, addr_size = self.implementation
        if addr_size == 0:
            raise MultiaddrValueError(f"Protocol admits no address value, but {repr(addr_value)} was passed.")
        if isinstance(addr_value, byteslike):
            assert raw_decoder is not None
            addr_value_str = raw_decoder(addr_value) # raises MultiaddrValueError if addr_value is invalid
            if not isinstance(addr_value, bytes):
                addr_value = bytes(addr_value)
            return addr_value_str, addr_value
        validate(addr_value, str)
        assert raw_encoder is not None
        addr_value_bytes = raw_encoder(addr_value) # raises MultiaddrValueError if addr_value is invalid
        return addr_value, addr_value_bytes

    def addr(self, value: Union[str, BytesLike]) -> "Addr":
        """
            Returns an address for this protocol.

            Example usage:

            >>> ip4.addr("192.168.1.1")
            Addr('ip4', '192.168.1.1')
            >>> ip4.addr(bytes([192, 168, 1, 1]))
            Addr('ip4', '192.168.1.1')

            The same address can be obtained with slash syntax:

            >>> ip4/"192.168.1.256"
            Addr('ip4', '192.168.1.256')
            >>> ip4/b'\\xc0\\xa8\\x01\\x01'
            Addr('ip4', '192.168.1.1')

        """
        return Addr(self, value)

    def __contains__(self, value: Union[str, BytesLike]) -> bool:
        return self.is_addr_valid(value)

    @overload
    def __truediv__(self, value: Union["Proto", "Addr", "Multiaddr"]) -> "Multiaddr":
        ...

    @overload
    def __truediv__(self, value: Union[int, str, BytesLike]) -> "Addr":
        ...

    def __truediv__(self, value: Union[int, str, BytesLike, "Proto", "Addr", "Multiaddr"]) -> Union["Addr", "Multiaddr"]:
        if isinstance(value, int):
            value = str(value)
        if isinstance(value, (str,)+byteslike):
            return self.addr(value)
        if isinstance(value, (Addr, Proto)):
            return Multiaddr(self, value)
        if isinstance(value, Multiaddr):
            return Multiaddr(self, *value)
        return NotImplemented

    def __str__(self) -> str:
        return f"/{self.name}"

    def __bytes__(self) -> bytes:
        if self.addr_size != 0:
            raise MultiaddrValueError("Missing address value for protocol, cannot compute bytes.")
        return varint.encode(self.code)

    def __repr__(self) -> str:
        return f"Proto({repr(self.name)})"

    @property
    def _as_tuple(self) -> Tuple[Type["Proto"], Multicodec]:
        return (Proto, self.codec)

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, Proto):
            return NotImplemented
        return self._as_tuple == other._as_tuple

class Addr:
    """
        Container class for a single protocol address in a `multiaddr <https://multiformats.io/multiaddr/>`_.

        >>> a = Addr('ip4', '192.168.1.1')
        >>> a
        Addr('ip4', '192.168.1.1')
        >>> str(a)
        '/ip4/192.168.1.1'

        The slash notation provides a more literate way to construct protocol addresses:

        >>> a = ip4/"192.168.1.1"
        >>> a
        Addr('ip4', '192.168.1.1')

        Bytes for protocol addresses are computed according to the `TLV multiaddr format <https://multiformats.io/multiaddr/>`_:

        >>> bytes(ip4/"127.0.0.1").hex()
        '047f000001'
        >>> varint.encode(ip4.code).hex()
        '04'
        >>> bytes([127, 0, 0, 1]).hex()
          '7f000001'

        :param proto: the protocol for this address (by name, code, multicodec or protocol object)
        :type proto: :obj:`str`, :obj:`int`, :class:`~multiformats.multicodec.Multicodec` or :class:`Proto`
        :param value: the address value (as a human-readable string or in its binary form)
        :type value: :obj:`str` or :obj:`~multiformats.varint.BytesLike`
    """

    _proto: Proto
    _value: str
    _value_bytes: bytes

    __slots__ = ("__weakref__", "_proto", "_value", "_value_bytes")

    def __new__(cls, proto: Union[str, int, Multicodec, Proto], value: Union[str, BytesLike]) -> "Addr":
        if not isinstance(proto, Proto):
            proto = Proto(proto)
        value, value_bytes = proto.validate(value)
        instance: Addr = super().__new__(cls)
        instance._proto = proto
        instance._value = value
        instance._value_bytes = value_bytes
        return instance

    @property
    def proto(self) -> Proto:
        """
            The address protocol.

            Example usage:

            >>> a = Addr('ip4', '192.168.1.1')
            >>> a.proto
            Proto('ip4')

        """
        return self._proto

    @property
    def value(self) -> str:
        """
            The address value, as a string.

            Example usage:

            >>> a = Addr('ip4', '192.168.1.1')
            >>> a.value
            '192.168.1.1'

        """
        return self._value

    @property
    def value_bytes(self) -> bytes:
        """
            The address value, as bytes.

            Example usage:

            >>> a = Addr('ip4', '192.168.1.1')
            >>> a.value_bytes
            b'\\xc0\\xa8\\x01\\x01'
            >>> list(a.value_bytes)
            [192, 168, 1, 1]

        """
        return self._value_bytes

    def __truediv__(self, other: Union[Proto, "Addr", "Multiaddr"]) -> "Multiaddr":
        if isinstance(other, (Addr, Proto)):
            return Multiaddr(self, other)
        if isinstance(other, Multiaddr):
            return Multiaddr(self, *other)
        return NotImplemented

    def __str__(self) -> str:
        return f"{str(self.proto)}/{self.value}"

    def __bytes__(self) -> bytes:
        proto = self.proto
        value_bytes = self.value_bytes
        if proto.addr_size is None:
            assert value_bytes is not None
            l = varint.encode(len(value_bytes))
            return proto.codec.wrap(l+value_bytes)
        return proto.codec.wrap(value_bytes)

    def __repr__(self) -> str:
        return f"Addr({repr(self.proto.name)}, {repr(self.value)})"

    @property
    def _as_tuple(self) -> Tuple[Type["Addr"], Proto, Optional[str]]:
        return (Addr, self.proto, self.value)

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, Addr):
            return NotImplemented
        return self._as_tuple == other._as_tuple


class Multiaddr(Sequence[Union[Addr, Proto]]):
    """
        Container class for a `multiaddr <https://multiformats.io/multiaddr/>`_.

        Example usage:

        >>> ip4 = multiaddr.proto("ip4")
        >>> udp = multiaddr.proto("udp")
        >>> quic = multiaddr.proto("quic")
        >>> ma = ip4/"127.0.0.1"/udp/9090/quic
        >>> ma
        Multiaddr(Addr('ip4', '127.0.0.1'), Addr('udp', '9090'), Proto('quic'))
        >>> str(ma)
        '/ip4/127.0.0.1/udp/9090/quic'

        Bytes for multiaddrs are computed according to the `(TLV)+ multiaddr format <https://multiformats.io/multiaddr/>`_:

        >>> bytes(ip4/"127.0.0.1").hex()
        '047f000001'
        >>> bytes(udp/9090).hex()
                  '91022382'
        >>> bytes(quic).hex()
                          'cc03'
        >>> bytes(ma).hex()
        '047f00000191022382cc03'

        :param addrs: a sequence of protocols (not requiring address) and protocol addresses
        :type addrs: sequence of :class:`Proto` or :class:`Addr`

        :raises ValueError: if a :class:`Proto` instance appears at a place other than the last for a protocol requiring an address
        :raises ValueError: if a procool name appears more than once
    """

    _addrs: Tuple[Union[Addr, Proto], ...]
    _proto_map: Dict[Proto, int]
    _is_incomplete: bool

    __slots__ = ("__weakref__", "_addrs", "_proto_map", "_is_incomplete")

    def __new__(cls, *addrs: Union[Addr, Proto]) -> "Multiaddr":
        l = len(addrs)
        is_incomplete = False
        proto_map: Dict[Proto, int] = {}
        for idx, addr in enumerate(addrs):
            if isinstance(addr, Proto):
                proto = addr
                if proto.addr_size != 0:
                    if idx == l-1:
                        is_incomplete = True
                    else:
                        raise MultiaddrValueError(f"Protocol {repr(proto.name)} expects an address, but is followed by another protocol instead.")
            else:
                validate(addr, Addr)
                proto = addr.proto
            if proto in proto_map:
                raise MultiaddrValueError(f"Protocol {repr(proto.name)} appears twice in multiaddr.")
            proto_map[proto] = idx
        instance: Multiaddr = super().__new__(cls)
        instance._addrs = addrs
        instance._proto_map = proto_map
        instance._is_incomplete = is_incomplete
        return instance

    @property
    def is_incomplete(self) -> bool:
        """
            Whether this multiaddress is incomplete, i.e. it still requires an address for
            the last protocol in the sequence.

            >>> ma = ip4/"127.0.0.1"/udp
            >>> ma.is_incomplete
            True
            >>> str(ma)
            '/ip4/127.0.0.1/udp'
            >>> ma2 = ma/9090
            >>> str(ma2)
            '/ip4/127.0.0.1/udp/9090'
            >>> ma2.is_incomplete
            False

            Incomplete multiaddrs don't admit a byte representation:

            >>> bytes(ma)
            MultiaddrValueError: Missing address value for last protocol, cannot compute bytes.
            >>> bytes(ma2).hex()
            '047f00000191022382'

        """
        return self._is_incomplete

    def index(self, value: Union[Addr, Proto], start: int = 0, stop: Optional[int] = None) -> int:
        """
            Returns the unique index at which a protocol/address appears in the multiaddress:

            >>> ma = ip4/"127.0.0.1"/udp/9090/quic
            >>> str(ma)
            '/ip4/127.0.0.1/udp/9090/quic'
            >>> udp in ma
            True
            >>> ma.index(udp)
            1
            >>> ma[ma.index(udp)]
            Addr('udp', '9090')
            >>> ip4/"127.0.0.1" in ma
            True
            >>> ma.index(ip4/"127.0.0.1" in ma)
            0

            :param value: the protocol or protocol address being looked for
            :type value: :class:`Proto` or :class:`Addr`
            :param start: the optional starting index (include) for the search range
            :type start: :obj:`int`, *optional*
            :param stop: the optional stoppping index (excluded) for the search range
            :type stop: :obj:`int` or :obj:`None`, *optional*

            >>> ip4/"127.0.0.1" in ma
            True
            >>> ma.index(ip4/"127.0.0.1")
            0
            >>> ma.index(ip4/"127.0.0.1", start=1)
            MultiaddrValueError: Address Addr('ip4', '127.0.0.1') does not appear in sub-multiaddr /udp/9090/quic of multiaddr /ip4/127.0.0.1/udp/9090/quic

            :raises ValueError: if the protocol/address does not appear:

            >>> ip6 = Proto("ip6")
            >>> ip6 in ma
            False
            >>> ma.index(ip6)
            MultiaddrValueError: Protocol 'ip6' does not appear in multiaddr /ip4/127.0.0.1/udp/9090/quic
            >>> ip4/"127.0.0.2" in ma
            False
            >>> ma.index(ip4/"127.0.0.2")
            MultiaddrValueError: Address Addr('ip4', '127.0.0.2') does not appear in multiaddr /ip4/127.0.0.1/udp/9090/quic

        """
        validate(start, int)
        if stop is None:
            stop = len(self)
        validate(stop, int)
        if isinstance(value, Proto):
            proto = value
        else:
            validate(value, Addr)
            proto = value.proto
        if proto not in self._proto_map:
            raise MultiaddrValueError(f"Protocol {repr(proto.name)} does not appear in multiaddr {str(self)}")
        idx = self._proto_map[proto]
        if isinstance(value, Addr):
            if self[idx] != value:
                raise MultiaddrValueError(f"Address {repr(value)} does not appear in multiaddr {str(self)}")
            if not start <= idx < stop:
                raise MultiaddrValueError(f"Address {repr(value)} does not appear in sub-multiaddr {str(self[start:stop])} "
                                 f"of multiaddr {str(self)}")
        return idx

    def __contains__(self, value: Any) -> bool:
        if isinstance(value, (Addr, Proto)):
            try:
                self.index(value)
                return True
            except MultiaddrValueError:
                return False
        return False

    def __len__(self) -> int:
        return len(self._addrs)

    def __iter__(self) -> Iterator[Union[Addr, Proto]]:
        return iter(self._addrs)

    @overload
    def __getitem__(self, idx: int) -> Union[Addr, Proto]:
        ...

    @overload
    def __getitem__(self, idx: slice) -> "Multiaddr":
        ...

    def __getitem__(self, idx: Union[int, slice]) -> Union[Addr, Proto, "Multiaddr"]:
        if isinstance(idx, int):
            return self._addrs[idx]
        validate(idx, slice)
        return Multiaddr(*self._addrs[idx])

    def __truediv__(self, other: Union[int, str, BytesLike, Addr, Proto, "Multiaddr"]) -> "Multiaddr":
        if isinstance(other, (int, str,)+byteslike):
            if not self.is_incomplete:
                raise MultiaddrValueError("Unexpected address value. Expected Proto, Addr or Multiaddr.")
            addrs = list(self)
            tail_proto = addrs[-1]
            assert isinstance(tail_proto, Proto)
            return Multiaddr(*islice(addrs, 0, len(addrs)-1), tail_proto/other)
        if isinstance(other, (Addr, Proto)):
            if self.is_incomplete:
                raise MultiaddrValueError("Expected address value (string or binary).")
            return Multiaddr(*self, other)
        if isinstance(other, Multiaddr):
            if self.is_incomplete:
                raise MultiaddrValueError("Expected address value (string or binary).")
            return Multiaddr(*self, *other)
        return NotImplemented

    def __str__(self) -> str:
        return "".join(str(a) for a in self)

    def __bytes__(self) -> bytes:
        if self.is_incomplete:
            raise MultiaddrValueError("Missing address value for last protocol, cannot compute bytes.")
        return bytes(chain.from_iterable(bytes(addr) for addr in self))

    def __repr__(self) -> str:
        return f"Multiaddr({', '.join(repr(a) for a in self)})"

    @property
    def _as_tuple(self) -> Tuple[Type["Multiaddr"], Tuple[Union[Addr, Proto], ...]]:
        return (Multiaddr, self._addrs)

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, Multiaddr):
            return NotImplemented
        return self._as_tuple == other._as_tuple


def proto(name: Union[str, int, Multicodec]) -> Proto:
    """
        Convenience function to construct a :class:`Proto` instance.

        Example usage:

        >>> ip4 = multiaddr.proto("ip4")
        >>> ip4
        Proto("ip4")

        :param name: the protocol name, multicodec code or multicodec object
        :type name: :obj:`str`, :obj:`int` or :class:`~multiformats.multicodec.Multicodec`

    """
    return Proto(name)

def parse(s: str, allow_incomplete: bool = False) -> Multiaddr:
    """
        Parses a multiaddr from its human-readable string representation.

        Example usage:

        >>> s = '/ip4/127.0.0.1/udp/9090/quic'
        >>> multiaddr.parse(s)
        Multiaddr(Addr('ip4', '127.0.0.1'), Addr('udp', '9090'), Proto('quic'))

        Example usage with incomplete multiaddr:

        >>> s = '/ip4/127.0.0.1/udp'
        >>> multiaddr.parse(s)
        MultiaddrValueError: Decoded multiaddr is incomplete: /ip4/127.0.0.1/udp
        >>> multiaddr.parse(s, allow_incomplete=True)
        Multiaddr(Addr('ip4', '127.0.0.1'), Proto('udp'))

        :param s: the multiaddress in human-readable string form
        :type s: :obj:`str`
        :param allow_incomplete: whether to allow incomplete multiaddresses
        :type allow_incomplete: :obj:`bool`

        :raises ValueError: if ``allow_incomplete`` is :obj:`False` and the parsed multiaddress is incomplete
    """
    validate(s, str)
    validate(allow_incomplete, bool)
    tokens = s.split("/")
    protocol: Optional[Proto] = None
    segments: List[Union[Addr, Proto]] = []
    for token in islice(tokens, 1, None):
        if protocol is None:
            protocol = Proto(token)
            if not protocol.admits_addr:
                segments.append(protocol)
                protocol = None
        else:
            segments.append(protocol/token)
            protocol = None
    if protocol is not None:
        segments.append(protocol)
    ma = Multiaddr(*segments)
    if ma.is_incomplete and not allow_incomplete:
        raise MultiaddrValueError(f"Decoded multiaddr is incomplete: {str(ma)}")
    return ma


def decode(b: BytesLike) -> Multiaddr:
    """
        Decodes a multiaddr from its `(TLV)+` binary encoding.

        Example usage:

        >>> b = bytes.fromhex('047f00000191022382cc03')
        >>> multiaddr.decode(b)
        Multiaddr(Addr('ip4', '127.0.0.1'), Addr('udp', '9090'), Proto('quic'))

        :param b: the multiaddress in binary form
        :type b: :obj:`~multiformats.varint.BytesLike`
    """
    validate(b, BytesLike)
    b = memoryview(b)
    segments: List[Union[Addr, Proto]] = []
    while len(b) > 0:
        code, _, b = multicodec.unwrap_raw(b)
        protocol = Proto(code)
        if protocol.admits_addr:
            addr_size = protocol.addr_size
            if addr_size is None:
                addr_size, _, b = varint.decode_raw(b)
            addr_value_bytes = bytes(b[:addr_size])
            b = b[addr_size:]
            segments.append(protocol/addr_value_bytes)
        else:
            segments.append(protocol)
    ma = Multiaddr(*segments)
    assert not ma.is_incomplete
    return ma



"""
    Implementation for the ``md5`` and ``ripemd`` hash functions,
    using `hashlib <https://docs.python.org/3/library/hashlib.html>`_ and `pycryptodomex <https://github.com/Legrandin/pycryptodome/>`_.
"""

import hashlib
from typing import Optional

from multiformats.varint import BytesLike
from .utils import Hashfun, validate_hashfun_args

def _md5(data: BytesLike, size: Optional[int] = None) -> bytes:
    validate_hashfun_args(data, size, 16)
    m: hashlib._Hash = hashlib.md5() # pylint: disable = no-member
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def _jit_register_md5(m, register) -> bool: # type: ignore
    if register is not None:
        register("md5", _md5, 16)
    return True

def _ripemd(digest_bits: int) -> Hashfun:
    assert digest_bits == 160, "Only 'ripemd-160' is currently supported."
    try:
        from Cryptodome.Hash import RIPEMD160 # pylint: disable = import-outside-toplevel
    except ImportError as e:
        raise ImportError("Module 'Cryptodome' must be installed to use the 'ripemd-160' hash function. Consider running 'pip install pycryptodomex'.") from e
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, digest_bits//8)
        m = RIPEMD160.new()
        m.update(data)
        d = m.digest()
        return d if size is None else d[:size]
    return hashfun

def _jit_register_ripemd(m, register) -> bool: # type: ignore
    digest_bits = int(m[1])
    if register is not None:
        register(f"ripemd-{digest_bits}", _ripemd(digest_bits), digest_bits//8)
    return True



# pylint: disable = missing-docstring

import pytest

from multiformats import varint, multibase, multicodec, multihash, CID

test_data = [
    b"",
    b"Hello world!",
    b"Hello world!"*1024
]

bases = ["base58btc", "base32", "base64pad"]
codecs = ["dag-pb", "raw"]
hashfuns = ["sha2-256", "sha3-512", "skein1024-1024"]

test_cases_v1 = [
    (base, 1, codec, hashfun)
    for base in bases
    for codec in codecs
    for hashfun in hashfuns
]
test_cases_v0 = [
    ("base58btc", 0, "dag-pb", "sha2-256")
]

test_cases = test_cases_v1+test_cases_v0

@pytest.mark.parametrize("base, v, codec, hashfun", test_cases)
@pytest.mark.parametrize("data", test_data)
def test_construct(data: bytes, base: str, v: int, codec: str, hashfun: str) -> None:
    mb = multibase.get(base)
    mc = multicodec.get(codec)
    mh = multihash.get(hashfun)
    digest = multihash.digest(data, hashfun)
    raw_digest = multihash.unwrap(digest)
    cids = [
        CID(base, v, codec, digest),
        CID(base, v, codec, (hashfun, raw_digest)),
        CID(mb, v, mc.code, digest),
        CID(mb, v, mc.code, (mh.code, raw_digest)),
        CID(mb, v, mc, digest),
        CID(mb, v, mc, (mh, raw_digest)),
    ]
    for cid in cids:
        assert cid.base == mb
        assert cid.version == v
        assert cid.codec == mc
        assert cid.hashfun == mh
        assert cid.digest == digest
        assert cid.raw_digest == raw_digest

@pytest.mark.parametrize("base, v, codec, hashfun", test_cases_v1)
@pytest.mark.parametrize("data", test_data)
def test_set(data: bytes, base: str, v: int, codec: str, hashfun: str) -> None:
    digest = multihash.digest(data, hashfun)
    cid = CID(base, v, codec, digest)
    for new_base in ["base58btc", "base32", "base32pad", "base64", "base64pad", "proquint"]:
        for new_codec in ["identity", "dag-pb", "dag-cbor", "raw"]:
            new_cid = cid.set(base=new_base, codec=new_codec)
            assert new_cid.base.name == new_base
            assert new_cid.version == v
            assert new_cid.codec.name == new_codec
            assert new_cid.hashfun.name == hashfun
            assert new_cid.digest == digest

@pytest.mark.parametrize("base, v, codec, hashfun", test_cases)
@pytest.mark.parametrize("data", test_data)
def test_encode_decode(data: bytes, base: str, v: int, codec: str, hashfun: str) -> None:
    mc = multicodec.get(codec)
    digest = multihash.digest(data, hashfun)
    cid = CID(base, v, codec, digest)
    s = cid.encode()
    b = bytes(cid)
    assert s == str(cid)
    if v == 1:
        assert b == varint.encode(v)+varint.encode(mc.code)+digest
    else:
        assert b == digest
    assert cid == CID.decode(s)
    assert cid.set(base="base58btc") == CID.decode(b)
    if v == 1:
        for other_base in ["base58btc", "base32", "base32pad", "base64", "base64pad", "proquint"]:
            assert cid.set(base=other_base) == CID.decode(cid.encode(other_base))

peer_id_test_cases = [
    ("30820122300d06092a864886f70d01010105000382010f003082010a02820101"
    "009a56a5c11e2705d0bfe0cd1fa66d5e519095cc741b62ed99ddf129c32e046e"
    "5ba3958bb8a068b05a95a6a0623cc3c889b1581793cd84a34cc2307e0dd74c70"
    "b4f230c74e5063ecd8e906d372be4eba13f47d04427a717ac78cb12b4b9c2ab5"
    "591f36f98021a70f84d782c36c51819054228ff35a45efa3f82b27849ec89036"
    "26b4a4c4b40f9f74b79caf55253687124c79cb10cd3bc73f0c44fbd341e5417d"
    "2e85e900d22849d2bc85ca6bf037f1f5b4f9759b4b6942fccdf1140b30ea7557"
    "87deb5c373c5953c14d64b523959a76a32a599903974a98cf38d4aaac7e359f8"
    "6b00a91dcf424bf794592139e7097d7e65889259227c07155770276b6eda4cec"
    "370203010001", "RSA"),
    ("1498b5467a63dffa2dc9d9e069caf075d16fc33fdd4c3b01bfadae6433767d93", "Ed25519"),
]

@pytest.mark.parametrize("pk_bytes_hex, algorithm", peer_id_test_cases)
def test_peer_id(pk_bytes_hex: str, algorithm: str) -> None:
    pk_bytes = bytes.fromhex(pk_bytes_hex)
    peer_id = CID.peer_id(pk_bytes_hex)
    assert peer_id == CID.peer_id(pk_bytes)
    assert peer_id.base.name == "base32"
    assert peer_id.version == 1
    assert peer_id.codec.name == "libp2p-key"
    if len(pk_bytes) <= 42:
        hashfun = "identity"
    else:
        hashfun = "sha2-256"
    assert peer_id.hashfun.name == hashfun
    assert peer_id.digest == multihash.digest(pk_bytes, hashfun)
